from typing import List
from math import trunc

from opentrons.drivers import utils
from opentrons.hardware_control.modules import ThermocyclerStep

from . import types as command_types


def magdeck_engage() -> command_types.MagdeckEngageCommand:
    text = "Engaging Magnetic Module"
    return {"name": command_types.MAGDECK_ENGAGE, "payload": {"text": text}}


def magdeck_disengage() -> command_types.MagdeckDisengageCommand:
    text = "Disengaging Magnetic Module"
    return {"name": command_types.MAGDECK_DISENGAGE, "payload": {"text": text}}


def magdeck_calibrate() -> command_types.MagdeckCalibrateCommand:
    text = "Calibrating Magnetic Module"
    return {"name": command_types.MAGDECK_CALIBRATE, "payload": {"text": text}}


def tempdeck_set_temp(celsius: float) -> command_types.TempdeckSetTempCommand:
    # TODO: SB 2022-06-03 Investigate if we should change this to match hs style below
    temp = round(float(celsius), utils.TEMPDECK_GCODE_ROUNDING_PRECISION)
    text = (
        f"Setting Temperature Module temperature "
        f"to {temp} °C (rounded off to nearest integer)"
    )
    return {
        "name": command_types.TEMPDECK_SET_TEMP,
        "payload": {"celsius": celsius, "text": text},
    }


def tempdeck_await_temp(celsius: float) -> command_types.TempdeckAwaitTempCommand:
    text = (
        "Waiting for Temperature Module to reach temperature "
        "{temp} °C (rounded off to nearest integer)".format(
            temp=round(float(celsius), utils.TEMPDECK_GCODE_ROUNDING_PRECISION)
        )
    )
    return {
        "name": command_types.TEMPDECK_AWAIT_TEMP,
        "payload": {"celsius": celsius, "text": text},
    }


def tempdeck_deactivate() -> command_types.TempdeckDeactivateCommand:
    text = "Deactivating Temperature Module"
    return {"name": command_types.TEMPDECK_DEACTIVATE, "payload": {"text": text}}


def thermocycler_open() -> command_types.ThermocyclerOpenCommand:
    text = "Opening Thermocycler lid"
    return {"name": command_types.THERMOCYCLER_OPEN, "payload": {"text": text}}


def thermocycler_set_block_temp(
    temperature: float, hold_time_seconds: float, hold_time_minutes: float
) -> command_types.ThermocyclerSetBlockTempCommand:
    temp = round(float(temperature), utils.TC_GCODE_ROUNDING_PRECISION)
    text = f"Setting Thermocycler well block temperature to {temp} °C"
    total_seconds = None
    # TODO: BC 2019-09-05 this time resolving logic is partially duplicated
    # in the thermocycler api class definition, with this command logger
    # implementation, there isn't a great way to avoid this, but it should
    # be consolidated as soon as an alternative to the publisher is settled on.
    if hold_time_seconds or hold_time_minutes:
        given_seconds = hold_time_seconds or 0
        given_minutes = hold_time_minutes or 0
        total_seconds = given_seconds + (given_minutes * 60)

        clean_seconds = total_seconds % 60
        clean_minutes = (total_seconds - clean_seconds) / 60
        text += " with a hold time of "
        if clean_minutes > 0:
            text += f"{clean_minutes} minutes and "
        text += f"{clean_seconds} seconds"
    return {
        "name": command_types.THERMOCYCLER_SET_BLOCK_TEMP,
        "payload": {
            "temperature": temperature,
            "hold_time": total_seconds,
            "text": text,
        },
    }


def thermocycler_execute_profile(
    steps: List[ThermocyclerStep], repetitions: int
) -> command_types.ThermocyclerExecuteProfileCommand:
    text = (
        f"Thermocycler starting {repetitions} repetitions"
        f" of cycle composed of the following steps: {steps}"
    )
    return {
        "name": command_types.THERMOCYCLER_EXECUTE_PROFILE,
        "payload": {"text": text, "steps": steps},
    }


def thermocycler_wait_for_hold() -> command_types.ThermocyclerWaitForHoldCommand:
    text = "Waiting for hold time duration"
    return {"name": command_types.THERMOCYCLER_WAIT_FOR_HOLD, "payload": {"text": text}}


def thermocycler_wait_for_temp() -> command_types.ThermocyclerWaitForTempCommand:
    text = "Waiting for Thermocycler to reach target"
    return {"name": command_types.THERMOCYCLER_WAIT_FOR_TEMP, "payload": {"text": text}}


def thermocycler_set_lid_temperature(
    temperature: float,
) -> command_types.ThermocyclerSetLidTempCommand:
    temp = round(float(temperature), utils.TC_GCODE_ROUNDING_PRECISION)
    text = f"Setting Thermocycler lid temperature to {temp} °C"
    return {"name": command_types.THERMOCYCLER_SET_LID_TEMP, "payload": {"text": text}}


def thermocycler_deactivate_lid() -> command_types.ThermocyclerDeactivateLidCommand:
    text = "Deactivating Thermocycler lid heating"
    return {
        "name": command_types.THERMOCYCLER_DEACTIVATE_LID,
        "payload": {"text": text},
    }


def thermocycler_deactivate_block() -> command_types.ThermocyclerDeactivateBlockCommand:
    text = "Deactivating Thermocycler well block heating"
    return {
        "name": command_types.THERMOCYCLER_DEACTIVATE_BLOCK,
        "payload": {"text": text},
    }


def thermocycler_deactivate() -> command_types.ThermocyclerDeactivateCommand:
    text = "Deactivating Thermocycler"
    return {"name": command_types.THERMOCYCLER_DEACTIVATE, "payload": {"text": text}}


def thermocycler_wait_for_lid_temp() -> command_types.ThermocyclerWaitForLidTempCommand:
    text = "Waiting for Thermocycler lid to reach target temperature"
    return {
        "name": command_types.THERMOCYCLER_WAIT_FOR_LID_TEMP,
        "payload": {"text": text},
    }


def thermocycler_close() -> command_types.ThermocyclerCloseCommand:
    text = "Closing Thermocycler lid"
    return {"name": command_types.THERMOCYCLER_CLOSE, "payload": {"text": text}}


def heater_shaker_set_target_temperature(
    celsius: float,
) -> command_types.HeaterShakerSetTargetTemperatureCommand:
    formatted_temp = trunc(celsius)
    text = f"Setting Target Temperature of Heater-Shaker to {formatted_temp} °C"
    return {
        "name": command_types.HEATER_SHAKER_SET_TARGET_TEMPERATURE,
        "payload": {"text": text},
    }


def heater_shaker_wait_for_temperature() -> command_types.HeaterShakerWaitForTemperatureCommand:
    text = "Waiting for Heater-Shaker to reach target temperature"
    return {
        "name": command_types.HEATER_SHAKER_WAIT_FOR_TEMPERATURE,
        "payload": {"text": text},
    }


def heater_shaker_set_and_wait_for_shake_speed(
    rpm: int,
) -> command_types.HeaterShakerSetAndWaitForShakeSpeedCommand:
    text = f"Setting Heater-Shaker to Shake at {rpm} RPM and waiting until reached"
    return {
        "name": command_types.HEATER_SHAKER_SET_AND_WAIT_FOR_SHAKE_SPEED,
        "payload": {"text": text},
    }


def heater_shaker_open_labware_latch() -> command_types.HeaterShakerOpenLabwareLatchCommand:
    text = "Unlatching labware on Heater-Shaker"
    return {
        "name": command_types.HEATER_SHAKER_OPEN_LABWARE_LATCH,
        "payload": {"text": text},
    }


def heater_shaker_close_labware_latch() -> command_types.HeaterShakerCloseLabwareLatchCommand:
    text = "Latching labware on Heater-Shaker"
    return {
        "name": command_types.HEATER_SHAKER_CLOSE_LABWARE_LATCH,
        "payload": {"text": text},
    }


def heater_shaker_deactivate_shaker() -> command_types.HeaterShakerDeactivateShakerCommand:
    text = "Deactivating Shaker"
    return {
        "name": command_types.HEATER_SHAKER_DEACTIVATE_SHAKER,
        "payload": {"text": text},
    }


def heater_shaker_deactivate_heater() -> command_types.HeaterShakerDeactivateHeaterCommand:
    text = "Deactivating Heater"
    return {
        "name": command_types.HEATER_SHAKER_DEACTIVATE_HEATER,
        "payload": {"text": text},
    }



"""Update position estimators payload, result, and implementaiton."""

from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, List, Type
from typing_extensions import Literal

from ...types import MotorAxis
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence
from ...resources import ensure_ot3_hardware

from opentrons.hardware_control import HardwareControlAPI

if TYPE_CHECKING:
    from ...execution import GantryMover


UnsafeEngageAxesCommandType = Literal["unsafe/engageAxes"]


class UnsafeEngageAxesParams(BaseModel):
    """Payload required for an UnsafeEngageAxes command."""

    axes: List[MotorAxis] = Field(..., description="The axes for which to enable.")


class UnsafeEngageAxesResult(BaseModel):
    """Result data from the execution of an UnsafeEngageAxes command."""


class UnsafeEngageAxesImplementation(
    AbstractCommandImpl[
        UnsafeEngageAxesParams,
        SuccessData[UnsafeEngageAxesResult],
    ]
):
    """Enable axes command implementation."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        gantry_mover: GantryMover,
        **kwargs: object,
    ) -> None:
        self._hardware_api = hardware_api
        self._gantry_mover = gantry_mover

    async def execute(
        self, params: UnsafeEngageAxesParams
    ) -> SuccessData[UnsafeEngageAxesResult]:
        """Enable exes."""
        ot3_hardware_api = ensure_ot3_hardware(self._hardware_api)
        await ot3_hardware_api.engage_axes(
            self._gantry_mover.motor_axes_to_present_hardware_axes(params.axes)
        )
        return SuccessData(
            public=UnsafeEngageAxesResult(),
        )


class UnsafeEngageAxes(
    BaseCommand[UnsafeEngageAxesParams, UnsafeEngageAxesResult, ErrorOccurrence]
):
    """UnsafeEngageAxes command model."""

    commandType: UnsafeEngageAxesCommandType = "unsafe/engageAxes"
    params: UnsafeEngageAxesParams
    result: Optional[UnsafeEngageAxesResult] = None

    _ImplementationCls: Type[
        UnsafeEngageAxesImplementation
    ] = UnsafeEngageAxesImplementation


class UnsafeEngageAxesCreate(BaseCommandCreate[UnsafeEngageAxesParams]):
    """UnsafeEngageAxes command request model."""

    commandType: UnsafeEngageAxesCommandType = "unsafe/engageAxes"
    params: UnsafeEngageAxesParams

    _CommandCls: Type[UnsafeEngageAxes] = UnsafeEngageAxes



import asyncio
import re
import base64
from typing import List, Optional

from opentrons.drivers.asyncio.communication.errors import NoResponse
from opentrons.drivers.command_builder import CommandBuilder
from opentrons.drivers.asyncio.communication import AsyncResponseSerialConnection

from .abstract import AbstractFlexStackerDriver
from .errors import StackerErrorCodes, MotorStallDetected
from .types import (
    GCODE,
    ActiveRange,
    LEDPattern,
    MeasurementKind,
    MoveResult,
    SpadMapID,
    StackerAxis,
    PlatformStatus,
    Direction,
    StackerInfo,
    HardwareRevision,
    MoveParams,
    AxisParams,
    LimitSwitchStatus,
    LEDColor,
    StallGuardParams,
    TOFConfiguration,
    TOFMeasurement,
    TOFMeasurementFrame,
    TOFMeasurementResult,
    TOFSensor,
    TOFSensorMode,
    TOFSensorState,
    TOFSensorStatus,
)
from .utils import validate_histogram_frame


FS_BAUDRATE = 115200
DEFAULT_FS_TIMEOUT = 1
FS_MOVE_TIMEOUT = 20
FS_TOF_TIMEOUT = 20
FS_ACK = "OK\n"
FS_ERROR_KEYWORD = "err"
FS_ASYNC_ERROR_ACK = "async"
DEFAULT_COMMAND_RETRIES = 0
GCODE_ROUNDING_PRECISION = 2

# LED animation range values
MIN_DURATION_MS = 25  # 25ms
MAX_DURATION_MS = 10000  # 10s
MAX_REPS = 10

# TOF Sensor
TOF_FRAME_RETRIES = 1
NUMBER_OF_BINS = 128

# Stallguard defaults
STALLGUARD_CONFIG = {
    StackerAxis.X: StallGuardParams(StackerAxis.X, True, 2),
    StackerAxis.Z: StallGuardParams(StackerAxis.Z, True, 2),
}

STACKER_MOTION_CONFIG = {
    StackerAxis.X: {
        "home": AxisParams(
            run_current=1.5,  # mAmps
            hold_current=0.75,
            move_params=MoveParams(
                max_speed=10.0,  # mm/s
                acceleration=100.0,  # mm/s^2
                max_speed_discont=40.0,  # mm/s
            ),
        ),
        "move": AxisParams(
            run_current=1.0,
            hold_current=0.75,
            move_params=MoveParams(
                max_speed=200.0,
                acceleration=1500.0,
                max_speed_discont=40.0,
            ),
        ),
    },
    StackerAxis.Z: {
        "home": AxisParams(
            run_current=1.5,
            hold_current=1.8,
            move_params=MoveParams(
                max_speed=10.0,
                acceleration=100.0,
                max_speed_discont=25.0,
            ),
        ),
        "move": AxisParams(
            run_current=1.5,
            hold_current=0.5,
            move_params=MoveParams(
                max_speed=150.0,
                acceleration=500.0,
                max_speed_discont=25.0,
            ),
        ),
    },
    StackerAxis.L: {
        "home": AxisParams(
            run_current=0.8,
            hold_current=0.15,
            move_params=MoveParams(
                max_speed=100.0,
                acceleration=800.0,
                max_speed_discont=40.0,
            ),
        ),
        "move": AxisParams(
            run_current=0.6,
            hold_current=0.15,
            move_params=MoveParams(
                max_speed=100.0,
                acceleration=800.0,
                max_speed_discont=40.0,
            ),
        ),
    },
}


class FlexStackerDriver(AbstractFlexStackerDriver):
    """FLEX Stacker driver."""

    @classmethod
    def parse_device_info(cls, response: str) -> StackerInfo:
        """Parse stacker info."""
        _RE = re.compile(
            f"^{GCODE.DEVICE_INFO} FW:(?P<fw>\\S+) HW:Opentrons-flex-stacker-(?P<hw>\\S+) SerialNo:(?P<sn>\\S+)$"
        )
        m = _RE.match(response)
        if not m:
            raise ValueError(f"Incorrect Response for device info: {response}")
        return StackerInfo(
            m.group("fw"), HardwareRevision(m.group("hw")), m.group("sn")
        )

    @classmethod
    def parse_reset_reason(cls, response: str) -> int:
        """Parse the reset reason"""
        _RE = re.compile(rf"^{GCODE.GET_RESET_REASON} R:(?P<R>\d)$")
        match = _RE.match(response)
        if not match:
            raise ValueError(f"Incorrect Response for reset reason: {response}")
        return int(match.group("R"))

    @classmethod
    def parse_limit_switch_status(cls, response: str) -> LimitSwitchStatus:
        """Parse limit switch statuses."""
        field_names = LimitSwitchStatus.get_fields()
        pattern = r"\s".join([rf"{name}:(?P<{name}>\d)" for name in field_names])
        _RE = re.compile(f"^{GCODE.GET_LIMIT_SWITCH} {pattern}$")
        m = _RE.match(response)
        if not m:
            raise ValueError(f"Incorrect Response for limit switch status: {response}")
        return LimitSwitchStatus(*(bool(int(m.group(name))) for name in field_names))

    @classmethod
    def parse_platform_sensor_status(cls, response: str) -> PlatformStatus:
        """Parse platform statuses."""
        field_names = PlatformStatus.get_fields()
        pattern = r"\s".join([rf"{name}:(?P<{name}>\d)" for name in field_names])
        _RE = re.compile(f"^{GCODE.GET_PLATFORM_SENSOR} {pattern}$")
        m = _RE.match(response)
        if not m:
            raise ValueError(f"Incorrect Response for platform status: {response}")
        return PlatformStatus(*(bool(int(m.group(name))) for name in field_names))

    @classmethod
    def parse_door_closed(cls, response: str) -> bool:
        """Parse door closed."""
        _RE = re.compile(rf"^{GCODE.GET_DOOR_SWITCH} D:(\d)$")
        match = _RE.match(response)
        if not match:
            raise ValueError(f"Incorrect Response for door closed: {response}")
        return bool(int(match.group(1)))

    @classmethod
    def parse_installation_detected(cls, response: str) -> bool:
        """Parse install detection."""
        _RE = re.compile(rf"^{GCODE.GET_INSTALL_DETECTED} I:(\d)$")
        match = _RE.match(response)
        if not match:
            raise ValueError(
                f"Incorrect Response for installation detected: {response}"
            )
        return bool(int(match.group(1)))

    @classmethod
    def parse_move_params(cls, response: str) -> MoveParams:
        """Parse move params."""
        field_names = MoveParams.get_fields()
        pattern = r"\s".join(rf"{f}:(?P<{f}>(\d*\.)?\d+)" for f in field_names)
        _RE = re.compile(rf"^{GCODE.GET_MOVE_PARAMS} M:([XZL]) {pattern}$")
        m = _RE.match(response)
        if not m:
            raise ValueError(f"Incorrect Response for move params: {response}")
        return MoveParams(
            max_speed=float(m.group("V")),
            acceleration=float(m.group("A")),
            max_speed_discont=float(m.group("D")),
        )

    @classmethod
    def parse_stallguard_params(cls, response: str) -> StallGuardParams:
        """Parse stallguard params."""
        pattern = r"(?P<M>[XZL]):(?P<E>\d) T:(?P<T>\d+)"
        _RE = re.compile(f"^{GCODE.GET_STALLGUARD_THRESHOLD} {pattern}$")
        m = _RE.match(response)
        if not m:
            raise ValueError(f"Incorrect Response for stallfguard params: {response}")
        return StallGuardParams(
            axis=StackerAxis(m.group("M")),
            enabled=bool(int(m.group("E"))),
            threshold=int(m.group("T")),
        )

    @classmethod
    def parse_get_motor_register(cls, response: str) -> int:
        """Parse get motor register value."""
        pattern = r"(?P<M>[XZL]):(?P<R>\d+) V:(?P<V>\d+)"
        _RE = re.compile(f"^{GCODE.GET_MOTOR_DRIVER_REGISTER} {pattern}$")
        m = _RE.match(response)
        if not m:
            raise ValueError(
                f"Incorrect Response for get motor driver register: {response}"
            )
        return int(m.group("V"))

    @classmethod
    def parse_get_tof_sensor_register(cls, response: str) -> int:
        """Parse get tof sensor register value."""
        pattern = r"(?P<S>[XZ]):(?P<R>\d+) V:(?P<V>\d+)"
        _RE = re.compile(f"^{GCODE.GET_TOF_DRIVER_REGISTER} {pattern}$")
        m = _RE.match(response)
        if not m:
            raise ValueError(
                f"Incorrect Response for get tof sensor driver register: {response}"
            )
        return int(m.group("V"))

    @classmethod
    def parse_tof_sensor_status(cls, response: str) -> TOFSensorStatus:
        """Parse get tof sensor status response."""
        pattern = r"(?P<S>[XZ]):(?P<O>\d) T:(?P<T>\d) M:(?P<M>\d)"
        _RE = re.compile(f"^{GCODE.GET_TOF_SENSOR_STATUS} {pattern}$")
        m = _RE.match(response)
        if not m:
            raise ValueError(
                f"Incorrect Response for get tof sensor status: {response}"
            )
        return TOFSensorStatus(
            sensor=TOFSensor(m.group("S")),
            state=TOFSensorState(int(m.group("T"))),
            mode=TOFSensorMode(int(m.group("M"))),
            ok=bool(int(m.group("O"))),
        )

    @classmethod
    def parse_manage_tof_measurement(cls, response: str) -> TOFMeasurement:
        """Parse manage tof measurement response."""
        pattern = r"(?P<S>[XZ]) K:(?P<K>\d+) C:(?P<C>\d) L:(?P<L>\d+)"
        _RE = re.compile(f"^{GCODE.MANAGE_TOF_MEASUREMENT} {pattern}$")
        m = _RE.match(response)
        if not m:
            raise ValueError(
                f"Incorrect Response for manage tof measurements: {response}"
            )
        return TOFMeasurement(
            sensor=TOFSensor(m.group("S")),
            kind=MeasurementKind(int(m.group("K"))),
            cancelled=bool(int(m.group("C"))),
            total_bytes=int(m.group("L")),
        )

    @classmethod
    def parse_get_tof_measurement(cls, response: str) -> TOFMeasurementFrame:
        """Parse get tof measurement response frame."""
        pattern = r"(?P<S>[XZ]) I:(?P<I>\d+) D:(?P<D>.+)"
        _RE = re.compile(f"^{GCODE.GET_TOF_MEASUREMENT} {pattern}$")
        m = _RE.match(response)
        if not m:
            raise ValueError(
                f"Incorrect Response for get tof measurement frame: {response}"
            )
        return TOFMeasurementFrame(
            sensor=TOFSensor(m.group("S")),
            frame_id=int(m.group("I")),
            data=base64.b64decode(m.group("D")),
        )

    @classmethod
    def parse_get_tof_configuration(cls, response: str) -> TOFConfiguration:
        """Parse get tof sensor configuration response."""
        pattern = r"(?P<S>[XZ]) I:(?P<I>\d+) A:(?P<A>\d+) K:(?P<K>\d+) P:(?P<P>\d+) H:(?P<H>\d)"
        _RE = re.compile(f"^{GCODE.GET_TOF_CONFIGURATION} {pattern}$")
        m = _RE.match(response)
        if not m:
            raise ValueError(
                f"Incorrect Response for get tof sensor configuration: {response}"
            )
        return TOFConfiguration(
            sensor=TOFSensor(m.group("S")),
            spad_map_id=SpadMapID(int(m.group("I"))),
            active_range=ActiveRange(int(m.group("A"))),
            kilo_iterations=int(m.group("K")),
            report_period_ms=int(m.group("P")),
            histogram_dump=bool(m.group("H")),
        )

    @classmethod
    def append_move_params(
        cls, command: CommandBuilder, params: MoveParams | None
    ) -> CommandBuilder:
        """Append move params."""
        if params is not None:
            command.add_float("V", params.max_speed, GCODE_ROUNDING_PRECISION)
            command.add_float("A", params.acceleration, GCODE_ROUNDING_PRECISION)
            command.add_float("D", params.max_speed_discont, GCODE_ROUNDING_PRECISION)
        return command

    @classmethod
    async def create(
        cls, port: str, loop: Optional[asyncio.AbstractEventLoop]
    ) -> "FlexStackerDriver":
        """Create a FLEX Stacker driver."""
        connection = await AsyncResponseSerialConnection.create(
            port=port,
            baud_rate=FS_BAUDRATE,
            timeout=DEFAULT_FS_TIMEOUT,
            number_of_retries=DEFAULT_COMMAND_RETRIES,
            ack=FS_ACK,
            loop=loop,
            error_keyword=FS_ERROR_KEYWORD,
            async_error_ack=FS_ASYNC_ERROR_ACK,
            reset_buffer_before_write=True,
            error_codes=StackerErrorCodes,
        )
        return cls(connection)

    def __init__(self, connection: AsyncResponseSerialConnection) -> None:
        """
        Constructor

        Args:
            connection: Connection to the FLEX Stacker
        """
        self._connection = connection

    async def connect(self) -> None:
        """Connect to stacker."""
        await self._connection.open()

    async def disconnect(self) -> None:
        """Disconnect from stacker."""
        await self._connection.close()

    async def is_connected(self) -> bool:
        """Check connection to stacker."""
        return await self._connection.is_open()

    async def get_device_info(self) -> StackerInfo:
        """Get Device Info."""
        response = await self._connection.send_command(
            GCODE.DEVICE_INFO.build_command()
        )
        device_info = self.parse_device_info(response)
        reason_resp = await self._connection.send_command(
            GCODE.GET_RESET_REASON.build_command()
        )
        reason = self.parse_reset_reason(reason_resp)
        device_info.rr = reason
        return device_info

    async def set_serial_number(self, sn: str) -> None:
        """Set Serial Number."""
        if not re.match(r"^FST[\w]{1}[\d]{2}[\d]{8}[\d]+$", sn):
            raise ValueError(
                f"Invalid serial number: ({sn}) expected format: FSTA1020250119001"
            )
        resp = await self._connection.send_command(
            GCODE.SET_SERIAL_NUMBER.build_command().add_element(sn)
        )
        if not re.match(rf"^{GCODE.SET_SERIAL_NUMBER}$", resp):
            raise ValueError(f"Incorrect Response for set serial number: {resp}")

    async def enable_motors(self, axis: List[StackerAxis]) -> None:
        """Enables the axis motor if present, disables it otherwise."""
        command = GCODE.ENABLE_MOTORS.build_command()
        for a in axis:
            command.add_element(a.name)
        resp = await self._connection.send_command(command)
        if not re.match(rf"^{GCODE.ENABLE_MOTORS}$", resp):
            raise ValueError(f"Incorrect Response for enable motors: {resp}")

    async def stop_motors(self) -> None:
        """Stop all motor movement."""
        resp = await self._connection.send_command(GCODE.STOP_MOTORS.build_command())
        if not re.match(rf"^{GCODE.STOP_MOTORS}$", resp):
            raise ValueError(f"Incorrect Response for stop motors: {resp}")

    async def set_run_current(self, axis: StackerAxis, current: float) -> None:
        """Set axis peak run current in amps."""
        resp = await self._connection.send_command(
            GCODE.SET_RUN_CURRENT.build_command().add_float(axis.name, current)
        )
        if not re.match(rf"^{GCODE.SET_RUN_CURRENT}$", resp):
            raise ValueError(f"Incorrect Response for set run current: {resp}")

    async def set_ihold_current(self, axis: StackerAxis, current: float) -> None:
        """Set axis hold current in amps."""
        resp = await self._connection.send_command(
            GCODE.SET_IHOLD_CURRENT.build_command().add_float(axis.name, current)
        )
        if not re.match(rf"^{GCODE.SET_IHOLD_CURRENT}$", resp):
            raise ValueError(f"Incorrect Response for set ihold current: {resp}")

    async def set_stallguard_threshold(
        self, axis: StackerAxis, enable: bool, threshold: int
    ) -> None:
        """Enables and sets the stallguard threshold for the given axis motor."""
        assert axis != StackerAxis.L, "Stallguard not supported for L axis"
        if not -64 < threshold < 63:
            raise ValueError(
                f"Threshold value ({threshold}) should be between -64 and 63."
            )
        resp = await self._connection.send_command(
            GCODE.SET_STALLGUARD.build_command()
            .add_int(axis.name, int(enable))
            .add_int("T", threshold)
        )
        if not re.match(rf"^{GCODE.SET_STALLGUARD}$", resp):
            raise ValueError(f"Incorrect Response for set stallguard threshold: {resp}")

    async def enable_tof_sensor(self, sensor: TOFSensor, enable: bool) -> None:
        """Enable or disable the TOF sensor."""
        # Enabling the TOF sensor takes a while, so give extra timeout.
        timeout = FS_TOF_TIMEOUT if enable else DEFAULT_FS_TIMEOUT
        resp = await self._connection.send_command(
            GCODE.ENABLE_TOF_SENSOR.build_command().add_int(sensor.name, int(enable)),
            timeout=timeout,
        )
        if not re.match(rf"^{GCODE.ENABLE_TOF_SENSOR}$", resp):
            raise ValueError(f"Incorrect Response for enable TOF sensor: {resp}")

    async def set_tof_configuration(
        self,
        sensor: TOFSensor,
        spad_map_id: SpadMapID,
        active_range: Optional[ActiveRange] = None,
        kilo_iterations: Optional[int] = None,
        report_period_ms: Optional[int] = None,
        histogram_dump: Optional[bool] = None,
    ) -> None:
        """Set the configuration of the TOF sensor.

        :param sensor: The TOF sensor to configure.
        :param spad_map_id: The pre-defined SPAD map which sets the fov and focus area (14 default).
        :active_range: The operating mode Short-range high-accuracy (default) or long range.
        :kilo_iterations: The Measurement iterations times 1024 (4000 default).
        :report_period_ms: The reporting period before each measurement (500 default).
        :histogram_dump: Enables/Disables histogram measurements (True default).
        :return: None
        """
        command = (
            GCODE.SET_TOF_CONFIGURATION.build_command()
            .add_element(sensor.name)
            .add_int("I", spad_map_id.value)
        )
        if active_range:
            command.add_int("A", active_range.value)
        if kilo_iterations:
            command.add_int("K", kilo_iterations)
        if report_period_ms:
            command.add_int("P", report_period_ms)
        if histogram_dump:
            command.add_int("H", int(histogram_dump))
        resp = await self._connection.send_command(command)
        if not re.match(rf"^{GCODE.SET_TOF_CONFIGURATION}$", resp):
            raise ValueError(f"Incorrect Response for set TOF configuration: {resp}")

    async def get_tof_configuration(self, sensor: TOFSensor) -> TOFConfiguration:
        """Get the configuration of the TOF sensor."""
        resp = await self._connection.send_command(
            GCODE.GET_TOF_CONFIGURATION.build_command().add_element(sensor.value)
        )
        return self.parse_get_tof_configuration(resp)

    async def manage_tof_measurement(
        self,
        sensor: TOFSensor,
        kind: MeasurementKind = MeasurementKind.HISTOGRAM,
        start: bool = True,
    ) -> TOFMeasurement:
        """Start or stop Measurements from the TOF sensor."""
        command = (
            GCODE.MANAGE_TOF_MEASUREMENT.build_command()
            .add_element(sensor.name)
            .add_int("K", kind.value)
        )
        if not start:
            command.add_element("C")
        resp = await self._connection.send_command(command)
        return self.parse_manage_tof_measurement(resp)

    async def _get_tof_histogram_frame(
        self, sensor: TOFSensor, resend: bool = False
    ) -> TOFMeasurementFrame:
        """Get the next measurement frame from TOF sensor or resend previous."""
        command = GCODE.GET_TOF_MEASUREMENT.build_command().add_element(sensor.name)
        if resend:
            command.add_element("R")
        resp = await self._connection.send_command(command)
        return self.parse_get_tof_measurement(resp)

    async def get_tof_histogram(self, sensor: TOFSensor) -> TOFMeasurementResult:
        """Get the full histogram measurement from the TOF sensor."""
        data = []
        data_len = 0
        next_frame_id = 0
        resend = False
        retries = TOF_FRAME_RETRIES

        # Cancel any ongoing measurements
        status = await self.get_tof_sensor_status(sensor)
        if status.state == TOFSensorState.MEASURING:
            await self.manage_tof_measurement(sensor, start=False)

        # Put the TOF sensor into histogram measurement mode
        start = await self.manage_tof_measurement(
            sensor, MeasurementKind.HISTOGRAM, start=True
        )
        # Request frames until the full histogram is transfered
        while data_len < start.total_bytes:
            try:
                # Request next histogram frame
                frame = await self._get_tof_histogram_frame(sensor, resend=resend)
                # Validate histogram frame
                validate_histogram_frame(frame.data, next_frame_id)
                # append frame data
                channel = frame.data[7:]
                data.append(channel)
                data_len += len(channel)
                assert (
                    not data_len > start.total_bytes
                ), f"Invalid number of bytes, expected {start.total_bytes} got {data_len}."
                retries = TOF_FRAME_RETRIES
                next_frame_id += 1
                resend = False
            except (ValueError, NoResponse):
                # There was a timeout or timing error, request previous frame.
                if retries <= 0:
                    # Cancel the measurement
                    await self.manage_tof_measurement(sensor, start.kind, start=False)
                    raise RuntimeError(f"Exceded frame {next_frame_id} retries")
                retries -= 1
                resend = True
                continue
            except Exception as e:
                # Cancel the histogram request
                await self.manage_tof_measurement(sensor, start.kind, start=False)
                raise RuntimeError(f"Could not get {sensor} Histogram", e)

        # Parse channel bin measurements from data
        #
        # There are 10 channels (0 - 9), each channel is comprised of 3 * 128 bytes.
        # The data starts with all the 128 LSB bytes, then all the 128 MID bytes, and
        # finally all the 128 MSB bytes for all the channels as seen below.
        #
        # ch0 lsb = data[0], mid = data[10], msb = data[20]
        # ch1 lsb = data[1], mid = data[11], msb = data[21]
        # ch2 lsb = data[2], mid = data[12], msb = data[22]
        # ...
        # ch9 lsb = data[9], mid = data[19], msb = data[29]
        #
        # Each channel has 128 bins comprised of the lsb, mid, and msb payload from above.
        # We can get the bins for each channel by combining the lsb, mid, and msb payload
        # for that specific channel.
        bins = {}
        for ch in range(10):  # 0-9 channels
            # Get the lsb, mid, and msb payload for the ch
            lsb = data[ch]
            mid = data[ch + 10]
            msb = data[ch + 20]
            # combine lsb, mid, and msb bytes to generate bin count for the ch
            bins[ch] = [
                (msb[b] << 16) | (mid[b] << 8) | lsb[b] for b in range(NUMBER_OF_BINS)
            ]
        return TOFMeasurementResult(start.sensor, start.kind, bins)

    async def set_motor_driver_register(
        self, axis: StackerAxis, reg: int, value: int
    ) -> None:
        """Set the register of the given motor axis driver to the given value."""
        resp = await self._connection.send_command(
            GCODE.SET_MOTOR_DRIVER_REGISTER.build_command()
            .add_int(axis.name, reg)
            .add_element(str(value))
        )
        if not re.match(rf"^{GCODE.SET_MOTOR_DRIVER_REGISTER}$", resp):
            raise ValueError(
                f"Incorrect Response for set motor driver register: {resp}"
            )

    async def get_motor_driver_register(self, axis: StackerAxis, reg: int) -> int:
        """Gets the register value of the given motor axis driver."""
        response = await self._connection.send_command(
            GCODE.GET_MOTOR_DRIVER_REGISTER.build_command().add_int(axis.name, reg)
        )
        return self.parse_get_motor_register(response)

    async def set_tof_driver_register(
        self, sensor: TOFSensor, reg: int, value: int
    ) -> None:
        """Set the register of the given tof sensor driver to the given value."""
        resp = await self._connection.send_command(
            GCODE.SET_TOF_DRIVER_REGISTER.build_command()
            .add_int(sensor.name, reg)
            .add_element(str(value))
        )
        if not re.match(rf"^{GCODE.SET_TOF_DRIVER_REGISTER}$", resp):
            raise ValueError(
                f"Incorrect Response for set tof sensor driver register: {resp}"
            )

    async def get_tof_driver_register(self, sensor: TOFSensor, reg: int) -> int:
        """Gets the register value of the given tof sensor driver."""
        response = await self._connection.send_command(
            GCODE.GET_TOF_DRIVER_REGISTER.build_command().add_int(sensor.name, reg)
        )
        return self.parse_get_tof_sensor_register(response)

    async def get_tof_sensor_status(self, sensor: TOFSensor) -> TOFSensorStatus:
        """Get the status of the tof sensor."""
        response = await self._connection.send_command(
            GCODE.GET_TOF_SENSOR_STATUS.build_command().add_element(sensor.name)
        )
        return self.parse_tof_sensor_status(response)

    async def get_motion_params(self, axis: StackerAxis) -> MoveParams:
        """Get the motion parameters used by the given axis motor."""
        response = await self._connection.send_command(
            GCODE.GET_MOVE_PARAMS.build_command().add_element(axis.name)
        )
        return self.parse_move_params(response)

    async def get_stallguard_threshold(self, axis: StackerAxis) -> StallGuardParams:
        """Get the stallguard parameters by the given axis motor."""
        response = await self._connection.send_command(
            GCODE.GET_STALLGUARD_THRESHOLD.build_command().add_element(axis.name)
        )
        return self.parse_stallguard_params(response)

    async def get_limit_switch(self, axis: StackerAxis, direction: Direction) -> bool:
        """Get limit switch status.

        :return: True if limit switch is triggered, False otherwise
        """
        response = await self.get_limit_switches_status()
        return response.get(axis, direction)

    async def get_limit_switches_status(self) -> LimitSwitchStatus:
        """Get limit switch statuses for all axes."""
        response = await self._connection.send_command(
            GCODE.GET_LIMIT_SWITCH.build_command()
        )
        return self.parse_limit_switch_status(response)

    async def get_platform_sensor(self, direction: Direction) -> bool:
        """Get platform sensor at one direction."""
        response = await self.get_platform_status()
        return response.get(direction)

    async def get_platform_status(self) -> PlatformStatus:
        """Get platform sensor status.

        :return: True if platform is detected, False otherwise
        """
        response = await self._connection.send_command(
            GCODE.GET_PLATFORM_SENSOR.build_command()
        )
        return self.parse_platform_sensor_status(response)

    async def get_hopper_door_closed(self) -> bool:
        """Get whether or not door is closed.

        :return: True if door is closed, False otherwise
        """
        response = await self._connection.send_command(
            GCODE.GET_DOOR_SWITCH.build_command()
        )
        return self.parse_door_closed(response)

    async def get_installation_detected(self) -> bool:
        """Get whether or not installation is detected.

        :return: True if installation is detected, False otherwise
        """
        response = await self._connection.send_command(
            GCODE.GET_INSTALL_DETECTED.build_command()
        )
        return self.parse_installation_detected(response)

    async def move_in_mm(
        self, axis: StackerAxis, distance: float, params: MoveParams | None = None
    ) -> MoveResult:
        """Move axis by the given distance in mm."""
        command = self.append_move_params(
            GCODE.MOVE_TO.build_command().add_float(
                axis.name, distance, GCODE_ROUNDING_PRECISION
            ),
            params,
        )
        try:
            resp = await self._connection.send_command(command, timeout=FS_MOVE_TIMEOUT)
            if not re.match(rf"^{GCODE.MOVE_TO}$", resp):
                raise ValueError(f"Incorrect Response for move to: {resp}")
        except MotorStallDetected:
            self.reset_serial_buffers()
            return MoveResult.STALL_ERROR
        return MoveResult.NO_ERROR

    async def move_to_limit_switch(
        self, axis: StackerAxis, direction: Direction, params: MoveParams | None = None
    ) -> MoveResult:
        """Move until limit switch is triggered."""
        command = self.append_move_params(
            GCODE.MOVE_TO_SWITCH.build_command().add_int(axis.name, direction.value),
            params,
        )
        try:
            resp = await self._connection.send_command(command, timeout=FS_MOVE_TIMEOUT)
            if not re.match(rf"^{GCODE.MOVE_TO_SWITCH}$", resp):
                raise ValueError(f"Incorrect Response for move to switch: {resp}")
        except MotorStallDetected:
            self.reset_serial_buffers()
            return MoveResult.STALL_ERROR
        return MoveResult.NO_ERROR

    async def home_axis(self, axis: StackerAxis, direction: Direction) -> MoveResult:
        """Home axis."""
        command = GCODE.HOME_AXIS.build_command().add_int(axis.name, direction.value)
        try:
            resp = await self._connection.send_command(command, timeout=FS_MOVE_TIMEOUT)
        except MotorStallDetected:
            self.reset_serial_buffers()
            return MoveResult.STALL_ERROR
        if not re.match(rf"^{GCODE.HOME_AXIS}$", resp):
            raise ValueError(f"Incorrect Response for home axis: {resp}")
        return MoveResult.NO_ERROR

    async def set_led(
        self,
        power: float,
        color: Optional[LEDColor] = None,
        external: Optional[bool] = None,
        pattern: Optional[LEDPattern] = None,
        duration: Optional[int] = None,
        reps: Optional[int] = None,
    ) -> None:
        """Set LED Status bar color and pattern.

        :param power: Power of the LED (0-1.0), 0 is off, 1 is full power
        :param color: Color of the LED
        :param external: True if external LED, False if internal LED
        :param pattern: Animation pattern of the LED status bar
        :param duration: Animation duration in milliseconds (25-10000), 10s max
        :param reps: Number of times to repeat the animation (-1 - 10), -1 is forever.
        """
        power = max(0, min(power, 1.0))
        command = GCODE.SET_LED.build_command().add_float(
            "P", power, GCODE_ROUNDING_PRECISION
        )
        if color is not None:
            command.add_int("C", color.value)
        if external is not None:
            command.add_int("K", int(external))
        if pattern is not None:
            command.add_int("A", pattern.value)
        if duration is not None:
            duration = max(MIN_DURATION_MS, min(duration, MAX_DURATION_MS))
            command.add_int("D", duration)
        if reps is not None:
            command.add_int("R", max(-1, min(reps, MAX_REPS)))
        resp = await self._connection.send_command(command)
        if not re.match(rf"^{GCODE.SET_LED}$", resp):
            raise ValueError(f"Incorrect Response for set led: {resp}")

    async def enter_programming_mode(self) -> None:
        """Reboot into programming mode"""
        command = GCODE.ENTER_BOOTLOADER.build_command()
        await self._connection.send_dfu_command(command)
        await self._connection.close()

    def reset_serial_buffers(self) -> None:
        """Reset the input and output serial buffers."""
        self._connection._serial.reset_input_buffer()
        self._connection._serial.reset_output_buffer()



"""Command models for Flex Stacker commands."""

from .store import (
    StoreCommandType,
    StoreParams,
    StoreResult,
    Store,
    StoreCreate,
)

from .retrieve import (
    RetrieveCommandType,
    RetrieveParams,
    RetrieveResult,
    Retrieve,
    RetrieveCreate,
)

from .set_stored_labware import (
    SetStoredLabwareCommandType,
    SetStoredLabwareParams,
    SetStoredLabwareResult,
    SetStoredLabware,
    SetStoredLabwareCreate,
    StackerStoredLabwareDetails,
)

from .fill import FillCommandType, FillParams, FillResult, Fill, FillCreate

from .empty import EmptyCommandType, EmptyParams, EmptyResult, Empty, EmptyCreate

from .close_latch import (
    CloseLatchCommandType,
    CloseLatchParams,
    CloseLatchResult,
    CloseLatch,
    CloseLatchCreate,
)
from .open_latch import (
    OpenLatchCommandType,
    OpenLatchParams,
    OpenLatchResult,
    OpenLatch,
    OpenLatchCreate,
)

from .prepare_shuttle import (
    PrepareShuttleCommandType,
    PrepareShuttleParams,
    PrepareShuttleResult,
    PrepareShuttle,
    PrepareShuttleCreate,
)


__all__ = [
    # flexStacker/store
    "StoreCommandType",
    "StoreParams",
    "StoreResult",
    "Store",
    "StoreCreate",
    # flexStacker/retrieve
    "RetrieveCommandType",
    "RetrieveParams",
    "RetrieveResult",
    "Retrieve",
    "RetrieveCreate",
    # flexStacker/setStoredLabware
    "SetStoredLabwareCommandType",
    "SetStoredLabwareParams",
    "SetStoredLabwareResult",
    "SetStoredLabware",
    "SetStoredLabwareCreate",
    "StackerStoredLabwareDetails",
    # flexStacker/fill
    "FillCommandType",
    "FillParams",
    "FillResult",
    "Fill",
    "FillCreate",
    # flexStacker/empty
    "EmptyCommandType",
    "EmptyParams",
    "EmptyResult",
    "Empty",
    "EmptyCreate",
    # flexStacker/closeLatch
    "CloseLatchCommandType",
    "CloseLatchParams",
    "CloseLatchResult",
    "CloseLatch",
    "CloseLatchCreate",
    # flexStacker/openLatch
    "OpenLatchCommandType",
    "OpenLatchParams",
    "OpenLatchResult",
    "OpenLatch",
    "OpenLatchCreate",
    # flexStacker/prepareShuttle
    "PrepareShuttleCommandType",
    "PrepareShuttleParams",
    "PrepareShuttleResult",
    "PrepareShuttle",
    "PrepareShuttleCreate",
]



import asyncio
import contextlib
from dataclasses import replace
from functools import partial
import logging
import pathlib
from collections import OrderedDict
from typing import (
    Callable,
    Dict,
    Union,
    List,
    Optional,
    Tuple,
    Sequence,
    Set,
    Any,
    TypeVar,
    Mapping,
    cast,
)

from opentrons_shared_data.errors.exceptions import (
    PositionUnknownError,
    UnsupportedHardwareCommand,
)
from opentrons_shared_data.pipette import (
    pipette_load_name_conversions as pipette_load_name,
)
from opentrons_shared_data.pipette.types import PipetteName
from opentrons_shared_data.robot.types import RobotType
from opentrons import types as top_types
from opentrons.config import robot_configs
from opentrons.config.types import RobotConfig, OT3Config
from opentrons.drivers.rpi_drivers.types import USBPort, PortGroup

from .util import use_or_initialize_loop, check_motion_bounds, ot2_axis_to_string
from .instruments.ot2.pipette import (
    generate_hardware_configs,
    load_from_config_and_check_skip,
)
from .backends import Controller, Simulator
from .execution_manager import ExecutionManagerProvider
from .pause_manager import PauseManager
from .module_control import AttachedModulesControl
from .types import (
    Axis,
    CriticalPoint,
    DoorState,
    DoorStateNotification,
    ErrorMessageNotification,
    HardwareEventHandler,
    HardwareAction,
    MotionChecks,
    PauseType,
    StatusBarState,
    EstopState,
    SubSystem,
    SubSystemState,
    HardwareFeatureFlags,
    TipScrapeType,
)
from . import modules
from .robot_calibration import (
    RobotCalibrationProvider,
    RobotCalibration,
)
from .protocols import HardwareControlInterface
from .instruments.ot2.pipette_handler import PipetteHandlerProvider
from .instruments.ot2.instrument_calibration import load_pipette_offset
from .motion_utilities import (
    target_position_from_absolute,
    target_position_from_relative,
    target_position_from_plunger,
    deck_from_machine,
    machine_from_deck,
)


mod_log = logging.getLogger(__name__)

AttachedModuleSpec = Dict[str, List[Union[str, Tuple[str, str]]]]


class API(
    ExecutionManagerProvider,
    RobotCalibrationProvider,
    PipetteHandlerProvider[top_types.Mount],
    # This MUST be kept last in the inheritance list so that it is
    # deprioritized in the method resolution order; otherwise, invocations
    # of methods that are present in the protocol will call the (empty,
    # do-nothing) methods in the protocol. This will happily make all the
    # tests fail.
    HardwareControlInterface[RobotCalibration, top_types.Mount, RobotConfig],
):
    """This API is the primary interface to the hardware controller.

    Because the hardware manager controls access to the system's hardware
    as a whole, it is designed as a class of which only one should be
    instantiated at a time. This class's methods should be the only method
    of external access to the hardware. Each method may be minimal - it may
    only delegate the call to another submodule of the hardware manager -
    but its purpose is to be gathered here to provide a single interface.

    This implements the protocols in opentrons.hardware_control.protocols,
    and longer method docstrings may be found there. Docstrings for the
    methods in this class only note where their behavior is different or
    extended from that described in the protocol.
    """

    CLS_LOG = mod_log.getChild("API")

    def __init__(
        self,
        backend: Union[Controller, Simulator],
        loop: asyncio.AbstractEventLoop,
        config: RobotConfig,
        feature_flags: Optional[HardwareFeatureFlags] = None,
    ) -> None:
        """Initialize an API instance.

        This should rarely be explicitly invoked by an external user; instead,
        one of the factory methods build_hardware_controller or
        build_hardware_simulator should be used.
        """
        self._log = self.CLS_LOG.getChild(str(id(self)))
        self._config = config
        self._backend = backend
        self._loop = loop
        # If no feature flag set is defined, we will use the default values
        self._feature_flags = feature_flags or HardwareFeatureFlags()

        self._callbacks: Set[HardwareEventHandler] = set()
        # {'X': 0.0, 'Y': 0.0, 'Z': 0.0, 'A': 0.0, 'B': 0.0, 'C': 0.0}
        self._current_position: Dict[Axis, float] = {}

        self._last_moved_mount: Optional[top_types.Mount] = None
        # The motion lock synchronizes calls to long-running physical tasks
        # involved in motion. This fixes issue where for instance a move()
        # or home() call is in flight and something else calls
        # current_position(), which will not be updated until the move() or
        # home() call succeeds or fails.
        self._motion_lock = asyncio.Lock()
        self._door_state = DoorState.CLOSED
        self._pause_manager = PauseManager()
        ExecutionManagerProvider.__init__(self, isinstance(backend, Simulator))
        RobotCalibrationProvider.__init__(self)
        PipetteHandlerProvider.__init__(
            self, {top_types.Mount.LEFT: None, top_types.Mount.RIGHT: None}
        )

    @property
    def door_state(self) -> DoorState:
        return self._door_state

    @door_state.setter
    def door_state(self, door_state: DoorState) -> None:
        self._door_state = door_state

    def _update_door_state(self, door_state: DoorState) -> None:
        mod_log.info(f"Updating the window switch status: {door_state}")
        self.door_state = door_state
        for cb in self._callbacks:
            hw_event = DoorStateNotification(new_state=door_state)
            try:
                cb(hw_event)
            except Exception:
                mod_log.exception("Errored during door state event callback")

    def _reset_last_mount(self) -> None:
        self._last_moved_mount = None

    def get_deck_from_machine(
        self, machine_pos: Dict[Axis, float]
    ) -> Dict[Axis, float]:
        return deck_from_machine(
            machine_pos=machine_pos,
            attitude=self._robot_calibration.deck_calibration.attitude,
            offset=top_types.Point(0, 0, 0),
            robot_type=cast(RobotType, "OT-2 Standard"),
        )

    @classmethod
    async def build_hardware_controller(  # noqa: C901
        cls,
        config: Union[RobotConfig, OT3Config, None] = None,
        port: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        firmware: Optional[Tuple[pathlib.Path, str]] = None,
        feature_flags: Optional[HardwareFeatureFlags] = None,
    ) -> "API":
        """Build a hardware controller that will actually talk to hardware.

        This method should not be used outside of a real robot, and on a
        real robot only one true hardware controller may be active at one
        time.

        :param config: A config to preload. If not specified, load the default.
        :param port: A port to connect to. If not specified, the default port
                     (found by scanning for connected FT232Rs).
        :param loop: An event loop to use. If not specified, use the result of
                     :py:meth:`asyncio.get_event_loop`.
        """
        checked_loop = use_or_initialize_loop(loop)
        if isinstance(config, RobotConfig):
            checked_config = config
        else:
            checked_config = robot_configs.load_ot2()
        backend = await Controller.build(checked_config)
        backend.set_lights(button=None, rails=False)

        async def blink() -> None:
            while True:
                backend.set_lights(button=True, rails=None)
                await asyncio.sleep(0.5)
                backend.set_lights(button=False, rails=None)
                await asyncio.sleep(0.5)

        blink_task = checked_loop.create_task(blink())
        try:
            try:
                await backend.connect(port)
                fw_version = backend.fw_version
            except Exception:
                mod_log.exception(
                    "Motor driver could not connect, reprogramming if possible"
                )
                fw_version = None

            if firmware is not None:
                if fw_version != firmware[1]:
                    await backend.update_firmware(str(firmware[0]), checked_loop, True)
                    await backend.connect(port)
            elif firmware is None and fw_version is None:
                msg = (
                    "Motor controller could not be connected and no "
                    "firmware was provided for (re)programming"
                )
                mod_log.error(msg)
                raise RuntimeError(msg)

            api_instance = cls(
                backend,
                loop=checked_loop,
                config=checked_config,
                feature_flags=feature_flags,
            )
            await api_instance.cache_instruments()
            module_controls = await AttachedModulesControl.build(
                api_instance, board_revision=backend.board_revision
            )
            backend.module_controls = module_controls
            checked_loop.create_task(backend.watch(loop=checked_loop))
            backend.start_gpio_door_watcher(
                loop=checked_loop, update_door_state=api_instance._update_door_state
            )
            return api_instance
        finally:
            blink_task.cancel()
            try:
                await blink_task
            except asyncio.CancelledError:
                pass

    @classmethod
    async def build_hardware_simulator(
        cls,
        attached_instruments: Optional[
            Dict[top_types.Mount, Dict[str, Optional[str]]]
        ] = None,
        attached_modules: Optional[Dict[str, List[modules.SimulatingModule]]] = None,
        config: Optional[Union[RobotConfig, OT3Config]] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        strict_attached_instruments: bool = True,
        feature_flags: Optional[HardwareFeatureFlags] = None,
    ) -> "API":
        """Build a simulating hardware controller.

        This method may be used both on a real robot and on dev machines.
        Multiple simulating hardware controllers may be active at one time.
        """

        if None is attached_instruments:
            attached_instruments = {}

        if None is attached_modules:
            attached_modules = {}

        checked_loop = use_or_initialize_loop(loop)
        if isinstance(config, RobotConfig):
            checked_config = config
        else:
            checked_config = robot_configs.load_ot2()
        backend = await Simulator.build(
            attached_instruments,
            attached_modules,
            checked_config,
            checked_loop,
            strict_attached_instruments,
        )
        api_instance = cls(
            backend,
            loop=checked_loop,
            config=checked_config,
            feature_flags=feature_flags,
        )
        await api_instance.cache_instruments()
        module_controls = await AttachedModulesControl.build(
            api_instance, board_revision=backend.board_revision
        )
        backend.module_controls = module_controls
        await backend.watch()
        return api_instance

    def __repr__(self) -> str:
        return "<{} using backend {}>".format(type(self), type(self._backend))

    async def get_serial_number(self) -> Optional[str]:
        return await self._backend.get_serial_number()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """The event loop used by this instance."""
        return self._loop

    @property
    def is_simulator(self) -> bool:
        """`True` if this is a simulator; `False` otherwise."""
        return isinstance(self._backend, Simulator)

    def register_callback(self, cb: HardwareEventHandler) -> Callable[[], None]:
        """Allows the caller to register a callback, and returns a closure
        that can be used to unregister the provided callback
        """
        self._callbacks.add(cb)

        def unregister() -> None:
            self._callbacks.remove(cb)

        return unregister

    def get_fw_version(self) -> str:
        """
        Return the firmware version of the connected motor control board.

        The version is a string retrieved directly from the attached hardware
        (or possibly simulator).
        """
        from_backend = self._backend.fw_version
        if from_backend is None:
            return "unknown"
        else:
            return from_backend

    @property
    def fw_version(self) -> str:
        return self.get_fw_version()

    @property
    def board_revision(self) -> str:
        return str(self._backend.board_revision)

    @property
    def attached_subsystems(self) -> Dict[SubSystem, SubSystemState]:
        return {}

    # Incidentals (i.e. not motion) API

    async def set_lights(
        self, button: Optional[bool] = None, rails: Optional[bool] = None
    ) -> None:
        """Control the robot lights."""
        self._backend.set_lights(button, rails)

    async def get_lights(self) -> Dict[str, bool]:
        """Return the current status of the robot lights.

        :returns: A dict of the lights: `{'button': bool, 'rails': bool}`
        """
        return self._backend.get_lights()

    async def identify(self, duration_s: int = 5) -> None:
        """Blink the button light to identify the robot."""
        count = duration_s * 4
        on = False
        for sec in range(count):
            then = self._loop.time()
            await self.set_lights(button=on)
            on = not on
            now = self._loop.time()
            await asyncio.sleep(max(0, 0.25 - (now - then)))
        await self.set_lights(button=True)

    async def set_status_bar_state(self, state: StatusBarState) -> None:
        """The status bar does not exist on OT-2!"""
        return None

    async def set_status_bar_enabled(self, enabled: bool) -> None:
        """The status bar does not exist on OT-2!"""
        return None

    def get_status_bar_state(self) -> StatusBarState:
        """There is no status bar on OT-2, return IDLE at all times."""
        return StatusBarState.IDLE

    @ExecutionManagerProvider.wait_for_running
    async def delay(self, duration_s: float) -> None:
        """Delay execution by pausing and sleeping."""
        self.pause(PauseType.DELAY)
        try:
            await self.do_delay(duration_s)
        finally:
            self.resume(PauseType.DELAY)

    @property
    def attached_modules(self) -> List[modules.AbstractModule]:
        return self._backend.module_controls.available_modules

    async def update_firmware(
        self,
        firmware_file: str,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        explicit_modeset: bool = True,
    ) -> str:
        """Update the firmware on the motor controller board.

        :param firmware_file: The path to the firmware file.
        :param explicit_modeset: `True` to force the smoothie into programming
                                 mode; `False` to assume it is already in
                                 programming mode.
        :param loop: An asyncio event loop to use; if not specified, the one
                     associated with this instance will be used.
        :returns: The stdout of the tool used to update the smoothie
        """
        if None is loop:
            checked_loop = self._loop
        else:
            checked_loop = loop
        return await self._backend.update_firmware(
            firmware_file, checked_loop, explicit_modeset
        )

    def has_gripper(self) -> bool:
        return False

    async def cache_instruments(
        self,
        require: Optional[Dict[top_types.Mount, PipetteName]] = None,
        skip_if_would_block: bool = False,
    ) -> None:
        """
        Scan the attached instruments, take necessary configuration actions,
        and set up hardware controller internal state if necessary.
        """
        self._log.info("Updating instrument model cache")
        checked_require = require or {}
        for mount, name in checked_require.items():
            if not pipette_load_name.supported_pipette(name):
                raise RuntimeError(f"{name} is not a valid pipette name")
        async with self._motion_lock:
            found = await self._backend.get_attached_instruments(checked_require)
        for mount, instrument_data in found.items():
            config = instrument_data.get("config")
            req_instr = checked_require.get(mount, None)
            pip_id = instrument_data.get("id")
            pip_offset_cal = load_pipette_offset(pip_id, mount)
            p, may_skip = load_from_config_and_check_skip(
                config,
                self._attached_instruments[mount],
                req_instr,
                pip_id,
                pip_offset_cal,
                self._feature_flags.use_old_aspiration_functions,
            )
            self._attached_instruments[mount] = p
            if req_instr and p:
                converted_name = pipette_load_name.convert_to_pipette_name_type(
                    req_instr
                )
                p.act_as(converted_name)

            if may_skip:
                self._log.info(f"Skipping configuration on {mount.name}")
                continue

            self._log.info(f"Doing full configuration on {mount.name}")
            hw_config = generate_hardware_configs(
                p, self._config, self._backend.board_revision
            )
            await self._backend.configure_mount(mount, hw_config)
        self._log.info("Instruments found: {}".format(self._attached_instruments))

    # Global actions API
    def pause(self, pause_type: PauseType) -> None:
        """
        Pause motion of the robot after a current motion concludes.
        """
        self._pause_manager.pause(pause_type)

        async def _chained_calls() -> None:
            await self._execution_manager.pause()
            self._backend.pause()

        asyncio.run_coroutine_threadsafe(_chained_calls(), self._loop)

    def pause_with_message(self, message: str) -> None:
        """As pause, but providing a message to registered callbacks."""
        self._log.warning(f"Pause with message: {message}")
        notification = ErrorMessageNotification(message=message)
        for cb in self._callbacks:
            cb(notification)
        self.pause(PauseType.PAUSE)

    def resume(self, pause_type: PauseType) -> None:
        """Resume motion after a call to pause."""
        self._pause_manager.resume(pause_type)

        if self._pause_manager.should_pause:
            return

        # Resume must be called immediately to awaken thread running hardware
        #  methods (ThreadManager)
        self._backend.resume()

        async def _chained_calls() -> None:
            # mirror what happens API.pause.
            await self._execution_manager.resume()
            self._backend.resume()

        asyncio.run_coroutine_threadsafe(_chained_calls(), self._loop)

    async def halt(self, disengage_before_stopping: bool = False) -> None:
        """Immediately stop motion, cancel execution manager and cancel running tasks.

        After this call, the smoothie will be in a bad state until a call to
        :py:meth:`stop`.
        """
        if disengage_before_stopping:
            await self._backend.hard_halt()
        await self._backend.halt()

    async def stop(self, home_after: bool = True) -> None:
        """
        Stop motion as soon as possible, reset, and optionally home.

        This will cancel motion (after the current call to :py:meth:`move`;
        see :py:meth:`pause` for more detail), then home and reset the
        robot. After this call, no further recovery is necessary.
        """
        await self._backend.halt()  # calls smoothie_driver.kill()
        await self.cancel_execution_and_running_tasks()
        self._log.info("Recovering from halt")
        await self.reset()
        await self.cache_instruments()

        if home_after:
            await self.home()

    def is_movement_execution_taskified(self) -> bool:
        return self.taskify_movement_execution

    def should_taskify_movement_execution(self, taskify: bool) -> None:
        self.taskify_movement_execution = taskify

    async def cancel_execution_and_running_tasks(self) -> None:
        await self._execution_manager.cancel()

    async def reset(self) -> None:
        """Reset the stored state of the system."""
        self._pause_manager.reset()
        await self._execution_manager.reset()
        await PipetteHandlerProvider.reset(self)

    # Gantry/frame (i.e. not pipette) action API
    async def home_z(
        self,
        mount: Optional[top_types.Mount] = None,
        allow_home_other: bool = True,
    ) -> None:
        """Home the Z-stage(s) of the instrument mounts.

        If given a mount, will try to only home that mount.
        However, if the other mount is currently extended,
        both mounts will be homed, unless `allow_home_other`
        is explicitly set to `False`.

        Setting `allow_home_other` to `False` is a bad idea,
        but the option exists for strict backwards compatibility.
        """
        if mount is not None and (
            self._last_moved_mount in [mount, None] or allow_home_other is False
        ):
            axes = [Axis.by_mount(mount)]
        else:
            axes = [Axis.Z, Axis.A]

        await self.home(axes)

    async def _do_plunger_home(
        self,
        axis: Optional[Axis] = None,
        mount: Optional[top_types.Mount] = None,
        acquire_lock: bool = True,
    ) -> None:
        assert (axis is not None) ^ (mount is not None), "specify either axis or mount"
        if axis:
            checked_axis = axis
            checked_mount = Axis.to_ot2_mount(checked_axis)
        if mount:
            checked_mount = mount
            checked_axis = Axis.of_plunger(checked_mount)
        instr = self.hardware_instruments[checked_mount]
        if not instr:
            return
        async with contextlib.AsyncExitStack() as stack:
            if acquire_lock:
                await stack.enter_async_context(self._motion_lock)
            with self._backend.save_current():
                self._backend.set_active_current(
                    {checked_axis: instr.plunger_motor_current.run}
                )
                await self._backend.home([ot2_axis_to_string(checked_axis)])
                # either we were passed False for our acquire_lock and we
                # should pass it on, or we acquired the lock above and
                # shouldn't do it again
                target_pos = target_position_from_plunger(
                    checked_mount,
                    instr.plunger_positions.bottom,
                    self._current_position,
                )
                await self._move(
                    target_pos,
                    acquire_lock=False,
                    home_flagged_axes=False,
                )

    @ExecutionManagerProvider.wait_for_running
    async def home_plunger(self, mount: top_types.Mount) -> None:
        """
        Home the plunger motor for a mount, and then return it to the 'bottom'
        position.
        """
        await self.current_position(mount=mount, refresh=True)
        await self._do_plunger_home(mount=mount, acquire_lock=True)

    @ExecutionManagerProvider.wait_for_running
    async def home(self, axes: Optional[List[Axis]] = None) -> None:
        """Home the entire robot and initialize current position."""
        # Should we assert/ raise an error or just remove non-ot2 axes and log warning?
        # No internal code passes OT3 axes as arguments on an OT2. But a user/ client
        # can still explicitly specify an OT3 axis even when working on an OT2.
        # Adding this check in order to prevent misuse of axes types.
        if axes:
            unsupported = list(axis not in Axis.ot2_axes() for axis in axes)
            if any(unsupported):
                raise UnsupportedHardwareCommand(
                    message=f"At least one axis in {axes} is not supported on the OT2.",
                    detail={"unsupported_axes": str(unsupported)},
                )
        self._reset_last_mount()
        # Initialize/update current_position
        checked_axes = axes or [ax for ax in Axis.ot2_axes()]
        gantry = [ax for ax in checked_axes if ax in Axis.gantry_axes()]
        smoothie_gantry = [ot2_axis_to_string(ax) for ax in gantry]
        smoothie_pos = {}
        plungers = [ax for ax in checked_axes if ax not in Axis.gantry_axes()]

        async with self._motion_lock:
            if smoothie_gantry:
                smoothie_pos.update(await self._backend.home(smoothie_gantry))
                self._current_position = self.get_deck_from_machine(
                    self._axis_map_from_string_map(smoothie_pos)
                )
            for plunger in plungers:
                await self._do_plunger_home(axis=plunger, acquire_lock=False)

    async def current_position(
        self,
        mount: top_types.Mount,
        critical_point: Optional[CriticalPoint] = None,
        refresh: bool = False,
        # TODO(mc, 2021-11-15): combine with `refresh` for more reliable
        # position reporting when motors are not homed
        fail_on_not_homed: bool = False,
    ) -> Dict[Axis, float]:
        """Return the postion (in deck coords) of the critical point of the
        specified mount.
        """
        z_ax = Axis.by_mount(mount)
        plunger_ax = Axis.of_plunger(mount)
        position_axes = [Axis.X, Axis.Y, z_ax, plunger_ax]

        if fail_on_not_homed:
            if not self._current_position:
                raise PositionUnknownError(
                    message=f"Current position of {str(mount)} pipette is unknown,"
                    " please home.",
                    detail={"mount": str(mount), "missing_axes": str(position_axes)},
                )
            axes_str = [ot2_axis_to_string(a) for a in position_axes]
            if not self._backend.is_homed(axes_str):
                unhomed = self._backend._unhomed_axes(axes_str)
                raise PositionUnknownError(
                    message=f"{str(mount)} pipette axes ({unhomed}) must be homed.",
                    detail={"mount": str(mount), "unhomed_axes": str(unhomed)},
                )
        elif not self._current_position and not refresh:
            raise PositionUnknownError(
                message="Current position is unknown; please home motors."
            )
        async with self._motion_lock:
            if refresh:
                smoothie_pos = await self._backend.update_position()
                self._current_position = self.get_deck_from_machine(
                    self._axis_map_from_string_map(smoothie_pos)
                )
            if mount == top_types.Mount.RIGHT:
                offset = top_types.Point(0, 0, 0)
            else:
                offset = top_types.Point(*self._config.left_mount_offset)

            cp = self.critical_point_for(mount, critical_point)
            return {
                Axis.X: self._current_position[Axis.X] + offset[0] + cp.x,
                Axis.Y: self._current_position[Axis.Y] + offset[1] + cp.y,
                z_ax: self._current_position[z_ax] + offset[2] + cp.z,
                plunger_ax: self._current_position[plunger_ax],
            }

    async def gantry_position(
        self,
        mount: top_types.Mount,
        critical_point: Optional[CriticalPoint] = None,
        refresh: bool = False,
        # TODO(mc, 2021-11-15): combine with `refresh` for more reliable
        # position reporting when motors are not homed
        fail_on_not_homed: bool = False,
    ) -> top_types.Point:
        """Return the position of the critical point for only gantry axes."""
        cur_pos = await self.current_position(
            mount,
            critical_point,
            refresh,
            fail_on_not_homed,
        )
        return top_types.Point(
            x=cur_pos[Axis.X], y=cur_pos[Axis.Y], z=cur_pos[Axis.by_mount(mount)]
        )

    # TODO(mc, 2022-05-13): return resulting gantry position
    async def move_to(
        self,
        mount: top_types.Mount,
        abs_position: top_types.Point,
        speed: Optional[float] = None,
        critical_point: Optional[CriticalPoint] = None,
        max_speeds: Optional[Dict[Axis, float]] = None,
    ) -> None:
        """
        Move the critical point of the specified mount to a location
        relative to the deck, at the specified speed.
        """
        if not self._current_position:
            await self.home()

        target_position = target_position_from_absolute(
            mount,
            abs_position,
            partial(self.critical_point_for, cp_override=critical_point),
            top_types.Point(*self._config.left_mount_offset),
            top_types.Point(0, 0, 0),
        )

        await self.prepare_for_mount_movement(mount)
        await self._move(target_position, speed=speed, max_speeds=max_speeds)

    async def move_axes(
        self,
        position: Mapping[Axis, float],
        speed: Optional[float] = None,
        max_speeds: Optional[Dict[Axis, float]] = None,
        expect_stalls: bool = False,
    ) -> None:
        """Moves the effectors of the specified axis to the specified position.
        The effector of the x,y axis is the center of the carriage.
        The effector of the pipette mount axis are the mount critical points but only in z.
        """
        raise UnsupportedHardwareCommand(
            message="move_axes is not supported on the OT-2.",
            detail={"axes_commanded": str(list(position.keys()))},
        )

    async def move_rel(
        self,
        mount: top_types.Mount,
        delta: top_types.Point,
        speed: Optional[float] = None,
        max_speeds: Optional[Dict[Axis, float]] = None,
        check_bounds: MotionChecks = MotionChecks.NONE,
        fail_on_not_homed: bool = False,
    ) -> None:
        """Move the critical point of the specified mount by a specified
        displacement in a specified direction, at the specified speed.
        """

        # TODO: Remove the fail_on_not_homed and make this the behavior all the time.
        # Having the optional arg makes the bug stick around in existing code and we
        # really want to fix it when we're not gearing up for a release.
        if not self._current_position:
            if fail_on_not_homed:
                raise PositionUnknownError(
                    message="Cannot make a relative move because absolute position"
                    " is unknown.",
                    detail={
                        "mount": str(mount),
                        "fail_on_not_homed": str(fail_on_not_homed),
                    },
                )
            else:
                await self.home()

        target_position = target_position_from_relative(
            mount, delta, self._current_position
        )

        axes_moving = [Axis.X, Axis.Y, Axis.by_mount(mount)]
        axes_str = [ot2_axis_to_string(a) for a in axes_moving]
        if fail_on_not_homed and not self._backend.is_homed(axes_str):
            unhomed = self._backend._unhomed_axes(axes_str)
            raise PositionUnknownError(
                message=f"{str(mount)} pipette axes ({unhomed}) must be homed.",
                detail={"mount": str(mount), "unhomed_axes": str(unhomed)},
            )

        await self.prepare_for_mount_movement(mount)
        await self._move(
            target_position,
            speed=speed,
            max_speeds=max_speeds,
            check_bounds=check_bounds,
        )

    async def _cache_and_maybe_retract_mount(self, mount: top_types.Mount) -> None:
        """Retract the 'other' mount if necessary

        If `mount` does not match the value in :py:attr:`_last_moved_mount`
        (and :py:attr:`_last_moved_mount` exists) then retract the mount
        in :py:attr:`_last_moved_mount`. Also unconditionally update
        :py:attr:`_last_moved_mount` to contain `mount`.
        """
        if mount != self._last_moved_mount and self._last_moved_mount:
            await self.retract(self._last_moved_mount, 10)
        self._last_moved_mount = mount

    async def prepare_for_mount_movement(self, mount: top_types.Mount) -> None:
        await self._cache_and_maybe_retract_mount(mount)

    @ExecutionManagerProvider.wait_for_running
    async def _move(
        self,
        target_position: "OrderedDict[Axis, float]",
        speed: Optional[float] = None,
        home_flagged_axes: bool = True,
        max_speeds: Optional[Dict[Axis, float]] = None,
        acquire_lock: bool = True,
        check_bounds: MotionChecks = MotionChecks.NONE,
    ) -> None:
        """Worker function to apply robot motion.

        Robot motion means the kind of motions that are relevant to the robot,
        i.e. only one pipette plunger and mount move at the same time, and an
        XYZ move in the coordinate frame of one of the pipettes.

        ``target_position`` should be an ordered dict (ordered by XYZABC)
        of deck calibrated values, containing any specified XY motion and
        at most one of a ZA or BC components. The frame in which to move
        is identified by the presence of (ZA) or (BC).
        """
        machine_pos = self._string_map_from_axis_map(
            machine_from_deck(
                deck_pos=target_position,
                attitude=self._robot_calibration.deck_calibration.attitude,
                offset=top_types.Point(0, 0, 0),
                robot_type=cast(RobotType, "OT-2 Standard"),
            )
        )

        bounds = self._backend.axis_bounds
        to_check = {
            ax: machine_pos[ot2_axis_to_string(ax)]
            for idx, ax in enumerate(target_position.keys())
            if ax in Axis.gantry_axes()
        }

        check_motion_bounds(to_check, target_position, bounds, check_bounds)
        checked_maxes = max_speeds or {}
        str_maxes = {ot2_axis_to_string(ax): val for ax, val in checked_maxes.items()}
        async with contextlib.AsyncExitStack() as stack:
            if acquire_lock:
                await stack.enter_async_context(self._motion_lock)
            try:
                await self._backend.move(
                    machine_pos,
                    speed=speed,
                    home_flagged_axes=home_flagged_axes,
                    axis_max_speeds=str_maxes,
                )
            except Exception:
                self._log.exception("Move failed")
                self._current_position.clear()
                raise
            else:
                self._current_position.update(target_position)

    def get_engaged_axes(self) -> Dict[Axis, bool]:
        """Which axes are engaged and holding."""
        return {Axis[ax]: eng for ax, eng in self._backend.engaged_axes().items()}

    @property
    def engaged_axes(self) -> Dict[Axis, bool]:
        return self.get_engaged_axes()

    async def disengage_axes(self, which: List[Axis]) -> None:
        await self._backend.disengage_axes([ot2_axis_to_string(ax) for ax in which])

    def axis_is_present(self, axis: Axis) -> bool:
        is_ot2 = axis in Axis.ot2_axes()
        if not is_ot2:
            return False
        if axis in Axis.pipette_axes():
            mount = Axis.to_ot2_mount(axis)
            if self.attached_pipettes.get(mount) is None:
                return False
        return True

    @ExecutionManagerProvider.wait_for_running
    async def _fast_home(self, axes: Sequence[str], margin: float) -> Dict[str, float]:
        converted_axes = "".join(axes)
        return await self._backend.fast_home(converted_axes, margin)

    async def retract(self, mount: top_types.Mount, margin: float = 10) -> None:
        """Pull the specified mount up to its home position.

        Works regardless of critical point or home status.
        """
        await self.retract_axis(Axis.by_mount(mount), margin)

    async def retract_axis(self, axis: Axis, margin: float = 10) -> None:
        """Pull the specified axis up to its home position.

        Works regardless of critical point or home status.
        """
        smoothie_ax = (ot2_axis_to_string(axis),)

        async with self._motion_lock:
            smoothie_pos = await self._fast_home(smoothie_ax, margin)
            self._current_position = self.get_deck_from_machine(
                self._axis_map_from_string_map(smoothie_pos)
            )

    # Gantry/frame (i.e. not pipette) config API
    @property
    def config(self) -> RobotConfig:
        """Get the robot's configuration object.

        :returns .RobotConfig: The object.
        """
        return self._config

    @config.setter
    def config(self, config: Union[RobotConfig, OT3Config]) -> None:
        """Replace the currently-loaded config"""
        if isinstance(config, RobotConfig):
            self._config = config
        else:
            self._log.error("Cannot use an OT-3 config on an OT-2")

    def get_config(self) -> RobotConfig:
        """
        Get the robot's configuration object.

        :returns .RobotConfig: The object.
        """
        return self.config

    def set_config(self, config: Union[OT3Config, RobotConfig]) -> None:
        """Replace the currently-loaded config"""
        if isinstance(config, RobotConfig):
            self.config = config
        else:
            self._log.error("Cannot use an OT-3 config on an OT-2")

    async def update_config(self, **kwargs: Any) -> None:
        """Update values of the robot's configuration.

        `kwargs` should contain keys of the robot's configuration. For
        instance, `update_config(log_level='debug)` would change the API
        server log level to :py:attr:`logging.DEBUG`.

        Documentation on keys can be found in the documentation for
        :py:class:`.RobotConfig`.
        """
        self._config = replace(self._config, **kwargs)

    @property
    def hardware_feature_flags(self) -> HardwareFeatureFlags:
        return self._feature_flags

    @hardware_feature_flags.setter
    def hardware_feature_flags(self, feature_flags: HardwareFeatureFlags) -> None:
        self._feature_flags = feature_flags

    async def update_deck_calibration(self, new_transform: RobotCalibration) -> None:
        pass

    # Pipette action API
    async def prepare_for_aspirate(
        self, mount: top_types.Mount, rate: float = 1.0
    ) -> None:
        """
        Prepare the pipette for aspiration.
        """
        instrument = self.get_pipette(mount)
        self.ready_for_tip_action(instrument, HardwareAction.PREPARE_ASPIRATE, mount)

        if instrument.current_volume == 0:
            speed = self.plunger_speed(
                instrument, instrument.blow_out_flow_rate, "aspirate"
            )
            bottom = instrument.plunger_positions.bottom
            target = target_position_from_plunger(mount, bottom, self._current_position)
            await self._move(
                target,
                speed=(speed * rate),
                home_flagged_axes=False,
            )
            instrument.ready_to_aspirate = True

    async def aspirate(
        self,
        mount: top_types.Mount,
        volume: Optional[float] = None,
        rate: float = 1.0,
        correction_volume: float = 0.0,
    ) -> None:
        """
        Aspirate a volume of liquid (in microliters/uL) using this pipette.
        """
        aspirate_spec = self.plan_check_aspirate(mount, volume, rate)
        if not aspirate_spec:
            return
        target_pos = target_position_from_plunger(
            mount,
            aspirate_spec.plunger_distance,
            self._current_position,
        )
        try:
            self._backend.set_active_current(
                {aspirate_spec.axis: aspirate_spec.current}
            )
            await self._move(
                target_pos,
                speed=aspirate_spec.speed,
                home_flagged_axes=False,
            )
        except Exception:
            self._log.exception("Aspirate failed")
            aspirate_spec.instr.set_current_volume(0)
            raise
        else:
            aspirate_spec.instr.add_current_volume(aspirate_spec.volume)

    async def dispense(
        self,
        mount: top_types.Mount,
        volume: Optional[float] = None,
        rate: float = 1.0,
        push_out: Optional[float] = None,
        correction_volume: float = 0.0,
        is_full_dispense: bool = False,
    ) -> None:
        """
        Dispense a volume of liquid in microliters(uL) using this pipette.
        """

        dispense_spec = self.plan_check_dispense(mount, volume, rate, push_out)
        if not dispense_spec:
            return
        target_pos = target_position_from_plunger(
            mount,
            dispense_spec.plunger_distance,
            self._current_position,
        )

        try:
            self._backend.set_active_current(
                {dispense_spec.axis: dispense_spec.current}
            )
            await self._move(
                target_pos,
                speed=dispense_spec.speed,
                home_flagged_axes=False,
            )
        except Exception:
            self._log.exception("Dispense failed")
            dispense_spec.instr.set_current_volume(0)
            raise
        else:
            dispense_spec.instr.remove_current_volume(dispense_spec.volume)

    async def blow_out(
        self, mount: top_types.Mount, volume: Optional[float] = None
    ) -> None:
        """
        Force any remaining liquid to dispense. The liquid will be dispensed at
        the current location of pipette
        """
        blowout_spec = self.plan_check_blow_out(mount)
        self._backend.set_active_current({blowout_spec.axis: blowout_spec.current})
        target_pos = target_position_from_plunger(
            mount,
            blowout_spec.plunger_distance,
            self._current_position,
        )

        try:
            await self._move(
                target_pos,
                speed=blowout_spec.speed,
                home_flagged_axes=False,
            )
        except Exception:
            self._log.exception("Blow out failed")
            raise
        finally:
            blowout_spec.instr.set_current_volume(0)
            blowout_spec.instr.ready_to_aspirate = False

    async def update_nozzle_configuration_for_mount(
        self,
        mount: top_types.Mount,
        back_left_nozzle: Optional[str],
        front_right_nozzle: Optional[str],
        starting_nozzle: Optional[str] = None,
    ) -> None:
        """
        Update a nozzle configuration for a given pipette.

        The expectation of this function is that the back_left_nozzle/front_right_nozzle are the two corners
        of a rectangle of nozzles. A call to this function that does not follow that schema will result
        in an error.

        :param mount: A robot mount that the instrument is on.
        :param back_left_nozzle: A string representing a nozzle name of the form <LETTER><NUMBER> such as 'A1'.
        :param front_right_nozzle: A string representing a nozzle name of the form <LETTER><NUMBER> such as 'A1'.
        :param starting_nozzle: A string representing the starting nozzle which will be used as the critical point
        of the pipette nozzle configuration. By default, the back left nozzle will be the starting nozzle if
        none is provided.
        :return: None.

        If none of the nozzle parameters are provided, the nozzle configuration will be reset to default.
        """
        if not back_left_nozzle and not front_right_nozzle and not starting_nozzle:
            await self.reset_nozzle_configuration(mount)
        else:
            assert back_left_nozzle and front_right_nozzle
            await self.update_nozzle_configuration(
                mount, back_left_nozzle, front_right_nozzle, starting_nozzle
            )

    async def tip_pickup_moves(
        self,
        mount: top_types.Mount,
        presses: Optional[int] = None,
        increment: Optional[float] = None,
    ) -> None:
        spec, _ = self.plan_check_pick_up_tip(
            mount=mount, presses=presses, increment=increment
        )
        self._backend.set_active_current(spec.plunger_currents)
        target_absolute = target_position_from_plunger(
            mount, spec.plunger_prep_pos, self._current_position
        )
        await self._move(
            target_absolute,
            home_flagged_axes=False,
        )

        for press in spec.presses:
            with self._backend.save_current():
                self._backend.set_active_current(press.current)
                target_down = target_position_from_relative(
                    mount, press.relative_down, self._current_position
                )
                await self._move(target_down, speed=press.speed)
            target_up = target_position_from_relative(
                mount, press.relative_up, self._current_position
            )
            await self._move(target_up)
        # neighboring tips tend to get stuck in the space between
        # the volume chamber and the drop-tip sleeve on p1000.
        # This extra shake ensures those tips are removed
        for rel_point, speed in spec.shake_off_list:
            await self.move_rel(mount, rel_point, speed=speed)

        await self.retract(mount, spec.retract_target)

    async def pick_up_tip(
        self,
        mount: top_types.Mount,
        tip_length: float,
        presses: Optional[int] = None,
        increment: Optional[float] = None,
        prep_after: bool = True,
    ) -> None:
        """
        Pick up tip from current location.
        """

        spec, _add_tip_to_instrs = self.plan_check_pick_up_tip(
            mount=mount, presses=presses, increment=increment, tip_length=tip_length
        )
        self._backend.set_active_current(spec.plunger_currents)
        target_absolute = target_position_from_plunger(
            mount, spec.plunger_prep_pos, self._current_position
        )
        await self._move(
            target_absolute,
            home_flagged_axes=False,
        )

        for press in spec.presses:
            with self._backend.save_current():
                self._backend.set_active_current(press.current)
                target_down = target_position_from_relative(
                    mount, press.relative_down, self._current_position
                )
                await self._move(target_down, speed=press.speed)
            target_up = target_position_from_relative(
                mount, press.relative_up, self._current_position
            )
            await self._move(target_up)
        # neighboring tips tend to get stuck in the space between
        # the volume chamber and the drop-tip sleeve on p1000.
        # This extra shake ensures those tips are removed
        for rel_point, speed in spec.shake_off_list:
            await self.move_rel(mount, rel_point, speed=speed)

        await self.retract(mount, spec.retract_target)
        _add_tip_to_instrs()

        if prep_after:
            await self.prepare_for_aspirate(mount)

    async def tip_drop_moves(
        self,
        mount: top_types.Mount,
        home_after: bool = True,
        ignore_plunger: bool = False,
        scrape_type: TipScrapeType = TipScrapeType.NONE,
    ) -> None:
        spec, _ = self.plan_check_drop_tip(mount, home_after)

        for move in spec.drop_moves:
            self._backend.set_active_current(move.current)
            target_pos = target_position_from_plunger(
                mount, move.target_position, self._current_position
            )
            await self._move(
                target_pos,
                speed=move.speed,
                home_flagged_axes=False,
            )
            if move.home_after:
                smoothie_pos = await self._fast_home(
                    axes=[ot2_axis_to_string(ax) for ax in move.home_axes],
                    margin=move.home_after_safety_margin,
                )
                self._current_position = self.get_deck_from_machine(
                    self._axis_map_from_string_map(smoothie_pos)
                )

        for shake in spec.shake_moves:
            await self.move_rel(mount, shake[0], speed=shake[1])

        self._backend.set_active_current(spec.ending_current)

    async def drop_tip(self, mount: top_types.Mount, home_after: bool = True) -> None:
        """Drop tip at the current location."""
        await self.tip_drop_moves(mount, home_after)

        # todo(mm, 2024-10-17): Ideally, callers would be able to replicate the behavior
        # of this method via self.drop_tip_moves() plus other public methods. This
        # currently prevents that: there is no public equivalent for
        # instrument.set_current_volume().
        instrument = self.get_pipette(mount)
        instrument.set_current_volume(0)

        self.set_current_tiprack_diameter(mount, 0.0)
        self.remove_tip(mount)

    async def create_simulating_module(
        self,
        model: modules.types.ModuleModel,
    ) -> modules.AbstractModule:
        """Get a simulating module hardware API interface for the given model."""
        assert (
            self.is_simulator
        ), "Cannot build simulating module from non-simulating hardware control API"

        return await self._backend.module_controls.build_module(
            port="",
            usb_port=USBPort(name="", port_number=1, port_group=PortGroup.MAIN),
            type=modules.ModuleType.from_model(model),
            sim_model=model.value,
        )

    def get_instrument_max_height(
        self, mount: top_types.Mount, critical_point: Optional[CriticalPoint] = None
    ) -> float:
        return PipetteHandlerProvider.instrument_max_height(
            self, mount, self._config.z_retract_distance, critical_point
        )

    async def clean_up(self) -> None:
        """Get the API ready to stop cleanly."""
        await self._backend.clean_up()

    MapPayload = TypeVar("MapPayload")

    @staticmethod
    def _axis_map_from_string_map(
        input_map: Dict[str, "API.MapPayload"]
    ) -> Dict[Axis, "API.MapPayload"]:
        return {Axis[k]: v for k, v in input_map.items()}

    def _string_map_from_axis_map(
        self, input_map: Dict[Axis, "API.MapPayload"]
    ) -> Dict[str, "API.MapPayload"]:
        return {ot2_axis_to_string(k): v for k, v in input_map.items()}

    def get_estop_state(self) -> EstopState:
        return EstopState.DISENGAGED



"""Blow-out command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal


from .pipetting_common import (
    OverpressureError,
    PipetteIdMixin,
    FlowRateMixin,
    blow_out_in_place,
)
from .movement_common import (
    WellLocationMixin,
    DestinationPositionResult,
    move_to_well,
    StallOrCollisionError,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..state.update_types import StateUpdate

from opentrons.hardware_control import HardwareControlAPI


if TYPE_CHECKING:
    from ..execution import MovementHandler, PipettingHandler
    from ..state.state import StateView
    from ..resources import ModelUtils


BlowOutCommandType = Literal["blowout"]


class BlowOutParams(PipetteIdMixin, FlowRateMixin, WellLocationMixin):
    """Payload required to blow-out a specific well."""

    pass


class BlowOutResult(DestinationPositionResult):
    """Result data from the execution of a blow-out command."""

    pass


_ExecuteReturn = Union[
    SuccessData[BlowOutResult],
    DefinedErrorData[OverpressureError] | DefinedErrorData[StallOrCollisionError],
]


class BlowOutImplementation(AbstractCommandImpl[BlowOutParams, _ExecuteReturn]):
    """BlowOut command implementation."""

    def __init__(
        self,
        movement: MovementHandler,
        pipetting: PipettingHandler,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._movement = movement
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api
        self._model_utils = model_utils

    async def execute(self, params: BlowOutParams) -> _ExecuteReturn:
        """Move to and blow-out the requested well."""
        move_result = await move_to_well(
            movement=self._movement,
            model_utils=self._model_utils,
            pipette_id=params.pipetteId,
            labware_id=params.labwareId,
            well_name=params.wellName,
            well_location=params.wellLocation,
        )
        if isinstance(move_result, DefinedErrorData):
            return move_result
        blow_out_result = await blow_out_in_place(
            pipette_id=params.pipetteId,
            flow_rate=params.flowRate,
            location_if_error={
                "retryLocation": (
                    move_result.public.position.x,
                    move_result.public.position.y,
                    move_result.public.position.z,
                )
            },
            pipetting=self._pipetting,
            model_utils=self._model_utils,
        )
        if isinstance(blow_out_result, DefinedErrorData):
            return DefinedErrorData(
                public=blow_out_result.public,
                state_update=StateUpdate.reduce(
                    move_result.state_update, blow_out_result.state_update
                ),
                state_update_if_false_positive=StateUpdate.reduce(
                    move_result.state_update,
                    blow_out_result.state_update_if_false_positive,
                ),
            )
        else:
            return SuccessData(
                public=BlowOutResult(position=move_result.public.position),
                state_update=StateUpdate.reduce(
                    move_result.state_update, blow_out_result.state_update
                ).set_pipette_ready_to_aspirate(
                    pipette_id=params.pipetteId, ready_to_aspirate=False
                ),
            )


class BlowOut(
    BaseCommand[
        BlowOutParams,
        BlowOutResult,
        OverpressureError | StallOrCollisionError,
    ]
):
    """Blow-out command model."""

    commandType: BlowOutCommandType = "blowout"
    params: BlowOutParams
    result: Optional[BlowOutResult] = None

    _ImplementationCls: Type[BlowOutImplementation] = BlowOutImplementation


class BlowOutCreate(BaseCommandCreate[BlowOutParams]):
    """Create blow-out command request model."""

    commandType: BlowOutCommandType = "blowout"
    params: BlowOutParams

    _CommandCls: Type[BlowOut] = BlowOut



from .ot3 import gripper_offset
from .ot2 import mark_bad_calibration

from .deck_configuration import (
    serialize_deck_configuration,
    deserialize_deck_configuration,
)

# TODO these functions are only used in robot server. We should think about moving them and/or
# abstracting it away from a robot specific function. We should also check if the tip rack
# definition information is still needed.
from .ot2.tip_length import (
    get_custom_tiprack_definition_for_tlc,
    get_all_tip_length_calibrations,
    _save_custom_tiprack_definition,
)
from .ot2.pipette_offset import get_all_pipette_offset_calibrations
from .ot3.deck_attitude import (
    save_robot_belt_attitude,
    get_robot_belt_attitude,
    delete_robot_belt_attitude,
)
from .ot2.deck_attitude import (
    save_robot_deck_attitude,
    get_robot_deck_attitude,
    delete_robot_deck_attitude,
)

from . import ot2, ot3, helpers

__all__ = [
    "helpers",
    # deck calibration functions
    "save_robot_deck_attitude",
    "get_robot_deck_attitude",
    "delete_robot_deck_attitude",
    "save_robot_belt_attitude",
    "get_robot_belt_attitude",
    "delete_robot_belt_attitude",
    # deck configuration functions
    "serialize_deck_configuration",
    "deserialize_deck_configuration",
    # functions only used in robot server
    "_save_custom_tiprack_definition",
    "get_custom_tiprack_definition_for_tlc",
    "get_all_pipette_offset_calibrations",
    "get_all_tip_length_calibrations",
    # file exports
    "gripper_offset",
    "mark_bad_calibration",
    # Platform specific submodules
    "ot2",
    "ot3",
]



from typing import Optional
import logging
import math

from opentrons.types import Point
from .instrument_calibration import (
    GripperCalibrationOffset,
    load_gripper_calibration_offset,
)
from opentrons.hardware_control.dev_types import GripperDict
from opentrons.hardware_control.types import (
    CriticalPoint,
    GripperJawState,
    GripperProbe,
)
from opentrons.hardware_control.errors import InvalidCriticalPoint
from opentrons_shared_data.errors.exceptions import (
    GripperNotPresentError,
    CommandPreconditionViolated,
)
from .gripper import Gripper


MOD_LOG = logging.getLogger(__name__)


class GripperHandler:
    GH_LOG = MOD_LOG.getChild("GripperHandler")

    def __init__(self, gripper: Optional[Gripper] = None):
        self._gripper = gripper
        self._log = self.GH_LOG.getChild(str(id(self)))

    def has_gripper(self) -> bool:
        return bool(self._gripper)

    def get_gripper(self) -> Gripper:
        gripper = self._gripper
        if not gripper:
            raise GripperNotPresentError(
                message="Cannot perform action without gripper attached"
            )
        return gripper

    def reset_gripper(self) -> None:
        """Reset the internal state of the gripper."""
        og_gripper = self._gripper
        if not og_gripper:
            return
        new_gripper = Gripper(
            og_gripper.config,
            load_gripper_calibration_offset(og_gripper.gripper_id),
            og_gripper.gripper_id,
            og_gripper._jaw_max_offset,
        )
        self._gripper = new_gripper

    async def reset(self) -> None:
        self._gripper = None

    @property
    def gripper(self) -> Optional[Gripper]:
        return self._gripper

    @gripper.setter
    def gripper(self, gripper: Optional[Gripper] = None) -> None:
        self._gripper = gripper

    def reset_instrument_offset(self, to_default: bool) -> None:
        """
        Temporarily reset the gripper offsets to default values.
        """
        gripper = self.get_gripper()
        gripper.reset_offset(to_default)

    def save_instrument_offset(self, delta: Point) -> GripperCalibrationOffset:
        """
        Save a new instrument offset.
        :param delta: The offset to set for the pipette.
        """
        gripper = self.get_gripper()
        self._log.info(f"Saving gripper {gripper.gripper_id} offset: {delta}")
        return gripper.save_offset(delta)

    def get_critical_point(self, cp_override: Optional[CriticalPoint] = None) -> Point:
        if not self._gripper:
            raise GripperNotPresentError()
        if cp_override == CriticalPoint.MOUNT:
            raise InvalidCriticalPoint(
                cp_override.name,
                "gripper",
                "The gripper mount may not be moved directly.",
            )
        return self._gripper.critical_point(cp_override)

    def get_gripper_dict(self) -> Optional[GripperDict]:
        if not self._gripper:
            # TODO (spp, 2023-04-19): Should this raise an error if fetching info of
            #  gripper that's not attached, like we do with pipettes?
            return None
        else:
            return self._gripper.as_dict()

    def get_attached_probe(self) -> Optional[GripperProbe]:
        return self.get_gripper().attached_probe

    def add_probe(self, probe: GripperProbe) -> None:
        """This is used for finding the critical point during calibration."""
        gripper = self.get_gripper()
        current_probe = self.get_attached_probe()
        if not current_probe:
            gripper.add_probe(probe)
        else:
            self._log.warning("add probe called with a probe already attached.")

    def remove_probe(self) -> None:
        gripper = self.get_gripper()
        current_probe = self.get_attached_probe()
        if current_probe:
            gripper.remove_probe()
        else:
            self._log.warning("remove probe called without a probe attached")

    def check_ready_for_calibration(self) -> None:
        """Raise an exception if a probe is not attached before calibration."""
        gripper = self.get_gripper()
        gripper.check_calibration_pin_location_is_accurate()

    def check_ready_for_jaw_move(self, command: str) -> None:
        """Raise an exception if it is not currently valid to move the jaw."""
        gripper = self.get_gripper()
        if gripper.state == GripperJawState.UNHOMED:
            raise CommandPreconditionViolated(
                message=f"Cannot {command} gripper jaw before homing",
                detail={
                    "command": command,
                    "jaw_state": str(gripper.state),
                },
            )

    def is_ready_for_idle(self) -> bool:
        """Gripper can idle when the jaw is not currently gripping."""
        gripper = self.get_gripper()
        if gripper.state == GripperJawState.UNHOMED:
            self._log.warning(
                "Gripper jaw is not homed and cannot move to idle position."
            )
            return False
        return gripper.state != GripperJawState.GRIPPING

    def is_ready_for_jaw_home(self) -> bool:
        """Raise an exception if it is not currently valid to home the jaw."""
        gripper = self.get_gripper()
        if gripper.state == GripperJawState.GRIPPING and not math.isclose(
            gripper.jaw_width, gripper.geometry.jaw_width["min"], abs_tol=5.0
        ):
            return False
        return True

    def set_jaw_state(self, state: GripperJawState) -> None:
        self.get_gripper().state = state

    def get_duty_cycle_by_grip_force(self, newton: float) -> float:
        gripper = self.get_gripper()
        return gripper.duty_cycle_by_force(newton)

    def set_jaw_displacement(self, mm: float) -> None:
        gripper = self.get_gripper()
        gripper.current_jaw_displacement = mm

    def is_valid_jaw_width(self, mm: float) -> bool:
        conf = self.get_gripper().geometry
        return conf.jaw_width["min"] <= mm <= conf.jaw_width["max"]



"""Command models to stop heating a Thermocycler's block."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


DeactivateBlockCommandType = Literal["thermocycler/deactivateBlock"]


class DeactivateBlockParams(BaseModel):
    """Input parameters to unset a Thermocycler's target block temperature."""

    moduleId: str = Field(..., description="Unique ID of the Thermocycler.")


class DeactivateBlockResult(BaseModel):
    """Result data from unsetting a Thermocycler's target block temperature."""


class DeactivateBlockImpl(
    AbstractCommandImpl[DeactivateBlockParams, SuccessData[DeactivateBlockResult]]
):
    """Execution implementation of a Thermocycler's deactivate block command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: DeactivateBlockParams
    ) -> SuccessData[DeactivateBlockResult]:
        """Unset a Thermocycler's target block temperature."""
        thermocycler_state = self._state_view.modules.get_thermocycler_module_substate(
            params.moduleId
        )
        thermocycler_hardware = self._equipment.get_module_hardware_api(
            thermocycler_state.module_id
        )

        if thermocycler_hardware is not None:
            await thermocycler_hardware.deactivate_block()

        return SuccessData(
            public=DeactivateBlockResult(),
        )


class DeactivateBlock(
    BaseCommand[DeactivateBlockParams, DeactivateBlockResult, ErrorOccurrence]
):
    """A command to unset a Thermocycler's target block temperature."""

    commandType: DeactivateBlockCommandType = "thermocycler/deactivateBlock"
    params: DeactivateBlockParams
    result: Optional[DeactivateBlockResult] = None

    _ImplementationCls: Type[DeactivateBlockImpl] = DeactivateBlockImpl


class DeactivateBlockCreate(BaseCommandCreate[DeactivateBlockParams]):
    """A request to create a Thermocycler's deactivate block command."""

    commandType: DeactivateBlockCommandType = "thermocycler/deactivateBlock"
    params: DeactivateBlockParams

    _CommandCls: Type[DeactivateBlock] = DeactivateBlock



import asyncio
from contextlib import asynccontextmanager
from typing import (
    Protocol,
    Dict,
    Optional,
    List,
    Mapping,
    AsyncIterator,
    Sequence,
    Tuple,
    Set,
    TypeVar,
)
from opentrons_shared_data.pipette.types import (
    PipetteName,
)
from opentrons.config.types import GantryLoad
from opentrons.hardware_control.types import (
    BoardRevision,
    Axis,
    OT3Mount,
    OT3AxisMap,
    InstrumentProbeType,
    MotorStatus,
    UpdateStatus,
    SubSystem,
    SubSystemState,
    TipStateType,
    GripperJawState,
    HardwareFeatureFlags,
    EstopOverallStatus,
    EstopState,
    HardwareEventHandler,
    HardwareEventUnsubscriber,
    HepaFanState,
    HepaUVState,
    StatusBarState,
    PipetteSensorResponseQueue,
)
from opentrons.hardware_control.module_control import AttachedModulesControl
from ..dev_types import OT3AttachedInstruments
from .types import HWStopCondition


Cls = TypeVar("Cls")


class FlexBackend(Protocol):
    """Flex backend mypy protocol."""

    async def get_serial_number(self) -> Optional[str]:
        ...

    @asynccontextmanager
    def restore_system_constraints(self) -> AsyncIterator[None]:
        ...

    @asynccontextmanager
    def grab_pressure(self, channels: int, mount: OT3Mount) -> AsyncIterator[None]:
        ...

    def set_pressure_sensor_available(
        self, pipette_axis: Axis, available: bool
    ) -> None:
        ...

    def get_pressure_sensor_available(self, pipette_axis: Axis) -> bool:
        ...

    def update_constraints_for_gantry_load(self, gantry_load: GantryLoad) -> None:
        ...

    def update_constraints_for_calibration_with_gantry_load(
        self,
        gantry_load: GantryLoad,
    ) -> None:
        ...

    def update_constraints_for_plunger_acceleration(
        self,
        mount: OT3Mount,
        acceleration: float,
        gantry_load: GantryLoad,
        high_speed_pipette: bool = False,
    ) -> None:
        ...

    @property
    def initialized(self) -> bool:
        """True when the hardware controller has initialized and is ready."""
        ...

    @initialized.setter
    def initialized(self, value: bool) -> None:
        ...

    @property
    def gear_motor_position(self) -> Optional[float]:
        ...

    @property
    def board_revision(self) -> BoardRevision:
        """Get the board revision"""
        ...

    @property
    def module_controls(self) -> AttachedModulesControl:
        """Get the module controls."""
        ...

    @module_controls.setter
    def module_controls(self, module_controls: AttachedModulesControl) -> None:
        """Set the module controls"""
        ...

    async def update_to_default_current_settings(self, gantry_load: GantryLoad) -> None:
        ...

    def update_feature_flags(self, feature_flags: HardwareFeatureFlags) -> None:
        """Update the hardware feature flags used by the hardware controller."""
        ...

    async def update_motor_status(self) -> None:
        """Retreieve motor and encoder status and position from all present devices"""
        ...

    async def update_motor_estimation(self, axes: Sequence[Axis]) -> None:
        """Update motor position estimation for commanded axes, and update cache of data."""
        # Simulate conditions as if there are no stalls, aka do nothing
        ...

    def _get_motor_status(
        self, axes: Sequence[Axis]
    ) -> Dict[Axis, Optional[MotorStatus]]:
        ...

    def get_invalid_motor_axes(self, axes: Sequence[Axis]) -> List[Axis]:
        """Get axes that currently do not have the motor-ok flag."""
        ...

    def get_invalid_encoder_axes(self, axes: Sequence[Axis]) -> List[Axis]:
        """Get axes that currently do not have the encoder-ok flag."""
        ...

    def check_motor_status(self, axes: Sequence[Axis]) -> bool:
        ...

    def check_encoder_status(self, axes: Sequence[Axis]) -> bool:
        ...

    async def update_position(self) -> OT3AxisMap[float]:
        """Get the current position."""
        ...

    async def update_encoder_position(self) -> OT3AxisMap[float]:
        """Get the encoder current position."""
        ...

    async def liquid_probe(
        self,
        mount: OT3Mount,
        max_p_distance: float,
        mount_speed: float,
        plunger_speed: float,
        threshold_pascals: float,
        plunger_impulse_time: float,
        num_baseline_reads: int,
        z_offset_for_plunger_prep: float,
        probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
        force_both_sensors: bool = False,
        response_queue: Optional[PipetteSensorResponseQueue] = None,
    ) -> float:
        ...

    async def move(
        self,
        origin: Dict[Axis, float],
        target: Dict[Axis, float],
        speed: float,
        stop_condition: HWStopCondition = HWStopCondition.none,
        nodes_in_moves_only: bool = True,
    ) -> None:
        """Move to a position.

        Args:
            target_position: Map of axis to position.
            home_flagged_axes: Whether to home afterwords.
            speed: Optional speed
            axis_max_speeds: Optional map of axis to speed.

        Returns:
            None
        """
        ...

    async def home(
        self, axes: Sequence[Axis], gantry_load: GantryLoad
    ) -> OT3AxisMap[float]:
        """Home axes.

        Args:
            axes: Optional list of axes.

        Returns:
            Homed position.
        """
        ...

    async def gripper_grip_jaw(
        self,
        duty_cycle: float,
        expected_displacement: float,
        stop_condition: HWStopCondition = HWStopCondition.none,
        stay_engaged: bool = True,
    ) -> None:
        """Move gripper inward."""
        ...

    async def gripper_home_jaw(self, duty_cycle: float) -> None:
        """Move gripper outward."""
        ...

    async def gripper_hold_jaw(
        self,
        encoder_position_um: int,
    ) -> None:
        ...

    async def get_jaw_state(self) -> GripperJawState:
        """Get the state of the gripper jaw."""
        ...

    async def tip_action(
        self, origin: float, targets: List[Tuple[float, float]]
    ) -> None:
        ...

    async def home_tip_motors(
        self,
        distance: float,
        velocity: float,
        back_off: bool = True,
    ) -> None:
        ...

    async def get_attached_instruments(
        self, expected: Mapping[OT3Mount, PipetteName]
    ) -> Mapping[OT3Mount, OT3AttachedInstruments]:
        """Get attached instruments.

        Args:
            expected: Which mounts are expected.

        Returns:
            A map of mount to pipette name.
        """
        ...

    async def get_limit_switches(self) -> OT3AxisMap[bool]:
        """Get the state of the gantry's limit switches on each axis."""
        ...

    async def set_active_current(self, axis_currents: OT3AxisMap[float]) -> None:
        """Set the active current.

        Args:
            axis_currents: Axes' currents

        Returns:
            None
        """
        ...

    @asynccontextmanager
    def motor_current(
        self,
        run_currents: Optional[OT3AxisMap[float]] = None,
        hold_currents: Optional[OT3AxisMap[float]] = None,
    ) -> AsyncIterator[None]:
        """Save the current."""
        ...

    @asynccontextmanager
    def restore_z_r_run_current(self) -> AsyncIterator[None]:
        """
        Temporarily restore the active current ONLY when homing or
        retracting the Z_R axis while the 96-channel is attached.
        """
        ...

    @asynccontextmanager
    def increase_z_l_hold_current(self) -> AsyncIterator[None]:
        """
        Temporarily increase the hold current when engaging the Z_L axis
        while the 96-channel is attached
        """
        ...

    async def watch(self, loop: asyncio.AbstractEventLoop) -> None:
        ...

    @property
    def axis_bounds(self) -> OT3AxisMap[Tuple[float, float]]:
        """Get the axis bounds."""
        ...

    @property
    def fw_version(self) -> Dict[SubSystem, int]:
        """Get the firmware version."""
        ...

    def axis_is_present(self, axis: Axis) -> bool:
        ...

    @property
    def update_required(self) -> bool:
        ...

    def update_firmware(
        self,
        subsystems: Set[SubSystem],
        force: bool = False,
    ) -> AsyncIterator[UpdateStatus]:
        """Updates the firmware on the OT3."""
        ...

    def engaged_axes(self) -> OT3AxisMap[bool]:
        """Get engaged axes."""
        ...

    async def update_engaged_axes(self) -> None:
        """Update engaged axes."""
        ...

    async def is_motor_engaged(self, axis: Axis) -> bool:
        """Check if axis is enabled."""
        ...

    async def disengage_axes(self, axes: List[Axis]) -> None:
        """Disengage axes."""
        ...

    async def engage_axes(self, axes: List[Axis]) -> None:
        """Engage axes."""
        ...

    async def set_lights(self, button: Optional[bool], rails: Optional[bool]) -> None:
        """Set the light states."""
        ...

    async def get_lights(self) -> Dict[str, bool]:
        """Get the light state."""
        ...

    def pause(self) -> None:
        """Pause the controller activity."""
        ...

    def resume(self) -> None:
        """Resume the controller activity."""
        ...

    async def halt(self) -> None:
        """Halt the motors."""
        ...

    async def probe(self, axis: Axis, distance: float) -> OT3AxisMap[float]:
        """Probe."""
        ...

    async def clean_up(self) -> None:
        """Clean up."""
        ...

    @staticmethod
    def home_position() -> OT3AxisMap[float]:
        ...

    async def capacitive_probe(
        self,
        mount: OT3Mount,
        moving: Axis,
        distance_mm: float,
        speed_mm_per_s: float,
        sensor_threshold_pf: float,
        probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
    ) -> bool:
        ...

    async def capacitive_pass(
        self,
        mount: OT3Mount,
        moving: Axis,
        distance_mm: float,
        speed_mm_per_s: float,
        probe: InstrumentProbeType,
    ) -> List[float]:
        ...

    @property
    def subsystems(self) -> Dict[SubSystem, SubSystemState]:
        ...

    async def get_tip_status(
        self, mount: OT3Mount, ht_operation_sensor: Optional[InstrumentProbeType] = None
    ) -> TipStateType:
        ...

    def current_tip_state(self, mount: OT3Mount) -> Optional[bool]:
        ...

    async def update_tip_detector(self, mount: OT3Mount, sensor_count: int) -> None:
        ...

    async def teardown_tip_detector(self, mount: OT3Mount) -> None:
        ...

    async def set_status_bar_state(self, state: StatusBarState) -> None:
        ...

    async def set_status_bar_enabled(self, enabled: bool) -> None:
        ...

    def get_status_bar_state(self) -> StatusBarState:
        ...

    @property
    def estop_status(self) -> EstopOverallStatus:
        ...

    def estop_acknowledge_and_clear(self) -> EstopOverallStatus:
        ...

    def get_estop_state(self) -> EstopState:
        ...

    def add_estop_callback(self, cb: HardwareEventHandler) -> HardwareEventUnsubscriber:
        ...

    def check_gripper_position_within_bounds(
        self,
        expected_grip_width: float,
        grip_width_uncertainty_wider: float,
        grip_width_uncertainty_narrower: float,
        jaw_width: float,
        max_allowed_grip_error: float,
        hard_limit_lower: float,
        hard_limit_upper: float,
    ) -> None:
        ...

    async def set_hepa_fan_state(self, fan_on: bool, duty_cycle: int) -> bool:
        """Sets the state and duty cycle of the Hepa/UV module."""
        ...

    async def get_hepa_fan_state(self) -> Optional[HepaFanState]:
        ...

    async def set_hepa_uv_state(self, light_on: bool, uv_duration_s: int) -> bool:
        """Sets the state and duration (seconds) of the UV light for the Hepa/UV module."""
        ...

    async def get_hepa_uv_state(self) -> Optional[HepaUVState]:
        ...

    async def increase_evo_disp_count(self, mount: OT3Mount) -> None:
        """Tell a pipette to increase it's evo-tip-dispense-count in eeprom."""
        ...

    async def read_env_temp_sensor(
        self, mount: OT3Mount, primary: bool
    ) -> Optional[float]:
        """Read and return the current sensor information."""
        ...

    async def read_env_hum_sensor(
        self, mount: OT3Mount, primary: bool
    ) -> Optional[float]:
        """Read and return the current sensor information."""
        ...

    async def read_pressure_sensor(
        self, mount: OT3Mount, primary: bool
    ) -> Optional[float]:
        """Read and return the current sensor information."""
        ...

    async def read_capacitive_sensor(
        self, mount: OT3Mount, primary: bool
    ) -> Optional[float]:
        """Read and return the current sensor information."""
        ...



"""Calibration Move To Maintenance Location command payload, result, and implementation models."""
from __future__ import annotations

import enum
from typing import TYPE_CHECKING, Type, Optional
from typing_extensions import Literal

from pydantic import BaseModel, Field

from opentrons.types import MountType, Point, Mount
from opentrons.hardware_control.types import Axis, CriticalPoint
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence
from opentrons.protocol_engine.resources.ot3_validation import ensure_ot3_hardware

if TYPE_CHECKING:
    from opentrons.hardware_control import HardwareControlAPI
    from ...state.state import StateView

# These offsets supplied from HW
_ATTACH_POINT = Point(x=0, y=110)
_MAX_Z_AXIS_MOTION_RANGE = 215
# These offsets are by eye measuring
_INSTRUMENT_ATTACH_Z_POINT = 400.0
_LEFT_MOUNT_Z_MARGIN = 5
# Move the right mount a bit higher than the left so the user won't forget to unscrew
_RIGHT_MOUNT_Z_MARGIN = 20

MoveToMaintenancePositionCommandType = Literal["calibration/moveToMaintenancePosition"]


class MaintenancePosition(enum.Enum):
    """Maintenance position options."""

    ATTACH_PLATE = "attachPlate"
    ATTACH_INSTRUMENT = "attachInstrument"


class MoveToMaintenancePositionParams(BaseModel):
    """Calibration set up position command parameters."""

    mount: MountType = Field(
        ...,
        description="Gantry mount to move maintenance position.",
    )

    maintenancePosition: MaintenancePosition = Field(
        MaintenancePosition.ATTACH_INSTRUMENT,
        description="The position the gantry mount needs to move to.",
    )


class MoveToMaintenancePositionResult(BaseModel):
    """Result data from the execution of a MoveToMaintenancePosition command."""


class MoveToMaintenancePositionImplementation(
    AbstractCommandImpl[
        MoveToMaintenancePositionParams,
        SuccessData[MoveToMaintenancePositionResult],
    ]
):
    """Calibration set up position command implementation."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        state_view: StateView,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._hardware_api = hardware_api

    async def execute(
        self, params: MoveToMaintenancePositionParams
    ) -> SuccessData[MoveToMaintenancePositionResult]:
        """Move the requested mount to a maintenance deck slot."""
        ot3_api = ensure_ot3_hardware(
            self._hardware_api,
        )
        max_height_z_tip = ot3_api.get_instrument_max_height(Mount.LEFT)
        # disengage the gripper z mount if present and enabled
        await ot3_api.prepare_for_mount_movement(Mount.LEFT)

        await ot3_api.retract(Mount.LEFT)
        current_pos = await ot3_api.gantry_position(
            Mount.LEFT, critical_point=CriticalPoint.MOUNT
        )
        await ot3_api.move_to(
            mount=Mount.LEFT,
            abs_position=Point(x=_ATTACH_POINT.x, y=_ATTACH_POINT.y, z=current_pos.z),
            critical_point=CriticalPoint.MOUNT,
        )

        if params.mount != MountType.EXTENSION:
            if params.maintenancePosition == MaintenancePosition.ATTACH_INSTRUMENT:
                mount = params.mount.to_hw_mount()
                mount_to_axis = Axis.by_mount(mount)
                await ot3_api.prepare_for_mount_movement(mount)
                await ot3_api.move_axes(
                    {
                        mount_to_axis: _INSTRUMENT_ATTACH_Z_POINT,
                    }
                )
                await ot3_api.disengage_axes([mount_to_axis])
            else:
                max_motion_range = max_height_z_tip - _MAX_Z_AXIS_MOTION_RANGE
                await ot3_api.move_axes(
                    {
                        Axis.Z_L: max_motion_range + _LEFT_MOUNT_Z_MARGIN,
                    }
                )
                await ot3_api.disengage_axes([Axis.Z_L])
                await ot3_api.move_axes(
                    {
                        Axis.Z_R: max_motion_range + _RIGHT_MOUNT_Z_MARGIN,
                    }
                )
                await ot3_api.disengage_axes([Axis.Z_R])

        return SuccessData(
            public=MoveToMaintenancePositionResult(),
        )


class MoveToMaintenancePosition(
    BaseCommand[
        MoveToMaintenancePositionParams,
        MoveToMaintenancePositionResult,
        ErrorOccurrence,
    ]
):
    """Calibration set up position command model."""

    commandType: MoveToMaintenancePositionCommandType = (
        "calibration/moveToMaintenancePosition"
    )
    params: MoveToMaintenancePositionParams
    result: Optional[MoveToMaintenancePositionResult] = None

    _ImplementationCls: Type[
        MoveToMaintenancePositionImplementation
    ] = MoveToMaintenancePositionImplementation


class MoveToMaintenancePositionCreate(
    BaseCommandCreate[MoveToMaintenancePositionParams]
):
    """Calibration set up position command creation request model."""

    commandType: MoveToMaintenancePositionCommandType = (
        "calibration/moveToMaintenancePosition"
    )
    params: MoveToMaintenancePositionParams

    _CommandCls: Type[MoveToMaintenancePosition] = MoveToMaintenancePosition



"""Protocol Engine types to do with instruments."""

from typing import Union

from dataclasses import dataclass

from pydantic import BaseModel

from opentrons.types import MountType
from opentrons_shared_data.pipette.types import PipetteNameType


class LoadedPipette(BaseModel):
    """A pipette that has been loaded."""

    id: str
    pipetteName: PipetteNameType
    mount: MountType


@dataclass(frozen=True)
class CurrentAddressableArea:
    """The latest addressable area the robot has accessed."""

    pipette_id: str
    addressable_area_name: str


@dataclass(frozen=True)
class CurrentWell:
    """The latest well that the robot has accessed."""

    pipette_id: str
    labware_id: str
    well_name: str


CurrentPipetteLocation = Union[CurrentWell, CurrentAddressableArea]


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class InstrumentOffsetVector(BaseModel):
    """Instrument Offset from home position to robot deck."""

    x: float
    y: float
    z: float



"""Protocol engine types to do with labware."""

from __future__ import annotations

from typing import Optional
from dataclasses import dataclass
from datetime import datetime

from pydantic import BaseModel, Field

from .location import LabwareLocation
from .labware_offset_location import (
    LegacyLabwareOffsetLocation,
    LabwareOffsetLocationSequence,
)
from .labware_offset_vector import LabwareOffsetVector
from .util import Vec3f


class OverlapOffset(Vec3f):
    """Offset representing overlap space of one labware on top of another labware or module."""


class LabwareOffset(BaseModel):
    """An offset that the robot adds to a pipette's position when it moves to a labware.

    During the run, if a labware is loaded whose definition URI and location
    both match what's found here, the given offset will be added to all
    pipette movements that use that labware as a reference point.
    """

    id: str = Field(..., description="Unique labware offset record identifier.")
    createdAt: datetime = Field(..., description="When this labware offset was added.")
    definitionUri: str = Field(..., description="The URI for the labware's definition.")
    location: LegacyLabwareOffsetLocation = Field(
        ...,
        deprecated=True,
        description="Where the labware is located on the robot. Deprecated and present only for backwards compatibility; cannot represent certain locations. Use locationSequence instead.",
    )
    locationSequence: Optional[LabwareOffsetLocationSequence] = Field(
        default=None,
        description="Where the labware is located on the robot. Can represent all locations, but may not be present for older runs.",
    )
    vector: LabwareOffsetVector = Field(
        ...,
        description="The offset applied to matching labware.",
    )


class LegacyLabwareOffsetCreate(BaseModel):
    """Create request data for a labware offset with a legacy location field."""

    definitionUri: str = Field(..., description="The URI for the labware's definition.")
    location: LegacyLabwareOffsetLocation = Field(
        ...,
        description="Where the labware is located on the robot.",
    )
    vector: LabwareOffsetVector = Field(
        ...,
        description="The offset applied to matching labware.",
    )


class LabwareOffsetCreate(BaseModel):
    """Create request data for a labware offset with a modern location sequence."""

    definitionUri: str = Field(..., description="The URI for the labware's definition.")
    locationSequence: LabwareOffsetLocationSequence = Field(
        ..., description="Where the labware is located on the robot."
    )
    vector: LabwareOffsetVector = Field(
        ..., description="The offset applied to matching labware."
    )


@dataclass(frozen=True)
class LabwareOffsetCreateInternal:
    """An internal-only labware offset creator that captures both old and new location arguments."""

    definitionUri: str
    locationSequence: LabwareOffsetLocationSequence
    legacyLocation: LegacyLabwareOffsetLocation
    vector: LabwareOffsetVector


class LoadedLabware(BaseModel):
    """A labware that has been loaded."""

    id: str
    loadName: str
    definitionUri: str
    location: LabwareLocation = Field(
        ..., description="The labware's current location."
    )
    lid_id: Optional[str] = Field(
        None,
        description=("Labware ID of a Lid currently loaded on top of the labware."),
    )
    offsetId: Optional[str] = Field(
        None,
        description=(
            "An ID referencing the labware offset"
            " that applies to this labware placement."
            " Null or undefined means no offset was provided for this load,"
            " so the default of (0, 0, 0) will be used."
        ),
    )
    displayName: Optional[str] = Field(
        None,
        description="A user-specified display name for this labware, if provided.",
    )



from typing import Optional, Dict, Sequence

from opentrons_shared_data.errors import GeneralError, EnumeratedError, ErrorCodes


class RuntimeParameterRequired(GeneralError):
    """An error raised when a parameter is required to be set for full analysis."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an RuntimeParameterRequired."""
        super().__init__(
            ErrorCodes.RUNTIME_PARAMETER_VALUE_REQUIRED, message, detail, wrapping
        )


class ParameterValueError(ValueError):
    """An error raised when a parameter value is not valid."""


class ParameterDefinitionError(ValueError):
    """An error raised when a parameter definition value is not valid."""


class ParameterNameError(ValueError):
    """An error raised when a parameter name or description is not valid."""


class IncompatibleParameterError(ValueError):
    """An error raised when a parameter conflicts with another parameter."""



"""Labware data resource provider.

This module is a wrapper around existing, but older, internal APIs to
abstract away rough edges until we can improve those underlying interfaces.
"""
import logging
from anyio import to_thread

from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition,
    LabwareDefinition3,
    labware_definition_type_adapter,
)

from opentrons.protocols.labware import get_labware_definition

# TODO (lc 09-26-2022) We should conditionally import ot2 or ot3 calibration
from opentrons.hardware_control.instruments.ot2 import (
    instrument_calibration as instr_cal,
)
from opentrons.calibration_storage.types import TipLengthCalNotFound


log = logging.getLogger(__name__)


class LabwareDataProvider:
    """Labware data provider."""

    @staticmethod
    async def get_labware_definition(
        load_name: str,
        namespace: str,
        version: int,
    ) -> LabwareDefinition:
        """Get a labware definition given the labware's identification.

        Note: this method hits the filesystem, which will have performance
        implications if it is called often.
        """
        return await to_thread.run_sync(
            LabwareDataProvider._get_labware_definition_sync,
            load_name,
            namespace,
            version,
        )

    @staticmethod
    def _get_labware_definition_sync(
        load_name: str, namespace: str, version: int
    ) -> LabwareDefinition:
        return labware_definition_type_adapter.validate_python(
            get_labware_definition(load_name, namespace, version)
        )

    @staticmethod
    async def get_calibrated_tip_length(
        pipette_serial: str,
        labware_definition: LabwareDefinition,
        nominal_fallback: float,
    ) -> float:
        """Get the calibrated tip length of a tip rack / pipette pair.

        Note: this method hits the filesystem, which will have performance
        implications if it is called often.
        """
        return await to_thread.run_sync(
            LabwareDataProvider._get_calibrated_tip_length_sync,
            pipette_serial,
            labware_definition,
            nominal_fallback,
        )

    @staticmethod
    def _get_calibrated_tip_length_sync(
        pipette_serial: str,
        labware_definition: LabwareDefinition,
        nominal_fallback: float,
    ) -> float:
        if isinstance(labware_definition, LabwareDefinition3):
            # FIXME(mm, 2025-02-19): This needs to be resolved for v8.4.0.
            # Tip length calibration internals don't yet support schema 3 because
            # it's probably an incompatible change at the filesystem level
            # (not downgrade-safe), and because robot-server's calibration sessions
            # are built atop opentrons.protocol_api.core.legacy, which does not (yet?)
            # support labware schema 3.
            # https://opentrons.atlassian.net/browse/EXEC-1230
            log.warning(
                f"Tip rack"
                f" {labware_definition.namespace}/{labware_definition.parameters.loadName}/{labware_definition.version}"
                f" has schema 3, so tip length calibration is currently unsupported."
                f" Using nominal fallback of {nominal_fallback}."
            )
            return nominal_fallback
        else:
            try:
                return instr_cal.load_tip_length_for_pipette(
                    pipette_serial, labware_definition
                ).tip_length
            except TipLengthCalNotFound as e:
                message = (
                    f"No calibrated tip length found for {pipette_serial},"
                    f" using nominal fallback value of {nominal_fallback}"
                )
                log.debug(message, exc_info=e)
                return nominal_fallback



"""Place labware payload, result, and implementaiton."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from opentrons_shared_data.labware.types import LabwareUri
from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from pydantic import BaseModel, Field

from opentrons.hardware_control.types import Axis, OT3Mount
from opentrons.motion_planning.waypoints import get_gripper_labware_placement_waypoints
from opentrons.protocol_engine.errors.exceptions import (
    CannotPerformGripperAction,
    GripperNotAttachedError,
)
from opentrons.types import Point

from ...types import (
    DeckSlotLocation,
    ModuleModel,
    OnDeckLabwareLocation,
)
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence
from ...resources import ensure_ot3_hardware

from opentrons.hardware_control import HardwareControlAPI, OT3HardwareControlAPI

if TYPE_CHECKING:
    from ...state.state import StateView
    from ...execution.equipment import EquipmentHandler


UnsafePlaceLabwareCommandType = Literal["unsafe/placeLabware"]


class UnsafePlaceLabwareParams(BaseModel):
    """Payload required for an UnsafePlaceLabware command."""

    labwareURI: str = Field(..., description="Labware URI for labware.")
    location: OnDeckLabwareLocation = Field(
        ..., description="Where to place the labware."
    )


class UnsafePlaceLabwareResult(BaseModel):
    """Result data from the execution of an UnsafePlaceLabware command."""


class UnsafePlaceLabwareImplementation(
    AbstractCommandImpl[
        UnsafePlaceLabwareParams,
        SuccessData[UnsafePlaceLabwareResult],
    ]
):
    """The UnsafePlaceLabware command implementation."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        state_view: StateView,
        equipment: EquipmentHandler,
        **kwargs: object,
    ) -> None:
        self._hardware_api = hardware_api
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: UnsafePlaceLabwareParams
    ) -> SuccessData[UnsafePlaceLabwareResult]:
        """Place Labware.

        This command is used only when the gripper is in the middle of moving
        labware but is interrupted before completing the move. (i.e., the e-stop
        is pressed, get into error recovery, etc).

        Unlike the `moveLabware` command, where you pick a source and destination
        location, this command takes the labwareURI of the labware to be moved
        and location to move it to.

        """
        ot3api = ensure_ot3_hardware(self._hardware_api)
        if not ot3api.has_gripper():
            raise GripperNotAttachedError("No gripper found to perform labware place.")

        if ot3api.gripper_jaw_can_home():
            raise CannotPerformGripperAction(
                "Cannot place labware when gripper is not gripping."
            )

        location = self._state_view.geometry.ensure_valid_gripper_location(
            params.location,
        )

        definition = self._state_view.labware.get_definition_by_uri(
            # todo(mm, 2024-11-07): This is an unsafe cast from untrusted input.
            # We need a str -> LabwareUri parse/validate function.
            LabwareUri(params.labwareURI)
        )

        # todo(mm, 2024-11-06): This is only correct in the special case of an
        # absorbance reader lid. Its definition currently puts the offsets for *itself*
        # in the property that's normally meant for offsets for its *children.*
        final_offsets = self._state_view.labware.get_child_gripper_offsets(
            labware_definition=definition, slot_name=None
        )
        drop_offset = (
            Point(
                final_offsets.dropOffset.x,
                final_offsets.dropOffset.y,
                final_offsets.dropOffset.z,
            )
            if final_offsets
            else None
        )

        if isinstance(params.location, DeckSlotLocation):
            self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
                params.location.slotName.id
            )

        # This is an absorbance reader, move the lid to its dock (staging area).
        if isinstance(location, DeckSlotLocation):
            module = self._state_view.modules.get_by_slot(location.slotName)
            if module and module.model == ModuleModel.ABSORBANCE_READER_V1:
                location = self._state_view.modules.absorbance_reader_dock_location(
                    module.id
                )

        # NOTE: When the estop is pressed, the gantry loses position, lets use
        # the encoders to sync position.
        # Ideally, we'd do a full home, but this command is used when
        # the gripper is holding the plate reader, and a full home would
        # bang it into the right window.
        await ot3api.home(axes=[Axis.Z_L, Axis.Z_R, Axis.Z_G])
        await ot3api.engage_axes([Axis.X, Axis.Y])
        await ot3api.update_axis_position_estimations([Axis.X, Axis.Y])

        # Place the labware down
        await self._start_movement(ot3api, definition, location, drop_offset)

        return SuccessData(public=UnsafePlaceLabwareResult())

    async def _start_movement(
        self,
        ot3api: OT3HardwareControlAPI,
        labware_definition: LabwareDefinition,
        location: OnDeckLabwareLocation,
        drop_offset: Optional[Point],
    ) -> None:
        gripper_homed_position = await ot3api.gantry_position(
            mount=OT3Mount.GRIPPER,
            refresh=True,
        )

        to_labware_center = self._state_view.geometry.get_labware_grip_point(
            labware_definition=labware_definition, location=location
        )

        movement_waypoints = get_gripper_labware_placement_waypoints(
            to_labware_center=to_labware_center,
            gripper_home_z=gripper_homed_position.z,
            drop_offset=drop_offset,
        )

        # start movement
        for waypoint_data in movement_waypoints:
            if waypoint_data.jaw_open:
                if waypoint_data.dropping:
                    # This `disengage_axes` step is important in order to engage
                    # the electronic brake on the Z axis of the gripper. The brake
                    # has a stronger holding force on the axis than the hold current,
                    # and prevents the axis from spuriously dropping when  e.g. the notch
                    # on the side of a falling tiprack catches the jaw.
                    await ot3api.disengage_axes([Axis.Z_G])
                await ot3api.ungrip()
                if waypoint_data.dropping:
                    # We lost the position estimation after disengaging the axis, so
                    # it is necessary to home it next
                    await ot3api.home_z(OT3Mount.GRIPPER)
            await ot3api.move_to(
                mount=OT3Mount.GRIPPER, abs_position=waypoint_data.position
            )


class UnsafePlaceLabware(
    BaseCommand[UnsafePlaceLabwareParams, UnsafePlaceLabwareResult, ErrorOccurrence]
):
    """UnsafePlaceLabware command model."""

    commandType: UnsafePlaceLabwareCommandType = "unsafe/placeLabware"
    params: UnsafePlaceLabwareParams
    result: Optional[UnsafePlaceLabwareResult] = None

    _ImplementationCls: Type[
        UnsafePlaceLabwareImplementation
    ] = UnsafePlaceLabwareImplementation


class UnsafePlaceLabwareCreate(BaseCommandCreate[UnsafePlaceLabwareParams]):
    """UnsafePlaceLabware command request model."""

    commandType: UnsafePlaceLabwareCommandType = "unsafe/placeLabware"
    params: UnsafePlaceLabwareParams

    _CommandCls: Type[UnsafePlaceLabware] = UnsafePlaceLabware



"""Protocol engine types to deal with locating things on the deck."""

from __future__ import annotations
from typing import Literal, Union, TypeGuard

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from opentrons.types import DeckSlotName, StagingSlotName


class DeckSlotLocation(BaseModel):
    """The location of something placed in a single deck slot."""

    slotName: DeckSlotName = Field(
        ...,
        description=(
            # This description should be kept in sync with LegacyLabwareOffsetLocation.slotName.
            "A slot on the robot's deck."
            "\n\n"
            'The plain numbers like `"5"` are for the OT-2,'
            ' and the coordinates like `"C2"` are for the Flex.'
            "\n\n"
            "When you provide one of these values, you can use either style."
            " It will automatically be converted to match the robot."
            "\n\n"
            "When one of these values is returned, it will always match the robot."
        ),
    )


class StagingSlotLocation(BaseModel):
    """The location of something placed in a single staging slot."""

    slotName: StagingSlotName = Field(
        ...,
        description=(
            # This description should be kept in sync with LegacyLabwareOffsetLocation.slotName.
            "A slot on the robot's staging area."
            "\n\n"
            "These apply only to the Flex. The OT-2 has no staging slots."
        ),
    )


class AddressableAreaLocation(BaseModel):
    """The location of something place in an addressable area. This is a superset of deck slots."""

    addressableAreaName: str = Field(
        ...,
        description=(
            "The name of the addressable area that you want to use."
            " Valid values are the `id`s of `addressableArea`s in the"
            " [deck definition](https://github.com/Opentrons/opentrons/tree/edge/shared-data/deck)."
        ),
    )


class ModuleLocation(BaseModel):
    """The location of something placed atop a hardware module."""

    moduleId: str = Field(
        ...,
        description="The ID of a loaded module from a prior `loadModule` command.",
    )


class OnLabwareLocation(BaseModel):
    """The location of something placed atop another labware."""

    labwareId: str = Field(
        ...,
        description="The ID of a loaded Labware from a prior `loadLabware` command.",
    )


class InStackerHopperLocation(BaseModel):
    """The location of a labware in a stacker hopper."""

    kind: Literal["inStackerHopper"] = "inStackerHopper"
    moduleId: str = Field(
        ..., description="The ID of the stacker in which this labware is."
    )


_OffDeckLocationType = Literal["offDeck"]
_SystemLocationType = Literal["systemLocation"]
OFF_DECK_LOCATION: _OffDeckLocationType = "offDeck"
SYSTEM_LOCATION: _SystemLocationType = "systemLocation"


def labware_location_is_off_deck(
    location: LabwareLocation,
) -> TypeGuard[_OffDeckLocationType]:
    """Check if a location is an off deck location."""
    return isinstance(location, str) and location == OFF_DECK_LOCATION


def labware_location_is_system(
    location: LabwareLocation,
) -> TypeGuard[_SystemLocationType]:
    """Check if a location is the system location."""
    return isinstance(location, str) and location == SYSTEM_LOCATION


class OnLabwareLocationSequenceComponent(BaseModel):
    """Labware on another labware."""

    kind: Literal["onLabware"] = "onLabware"
    labwareId: str
    lidId: str | SkipJsonSchema[None] = Field(None)


class OnModuleLocationSequenceComponent(BaseModel):
    """Labware on a module."""

    kind: Literal["onModule"] = "onModule"
    moduleId: str


class OnAddressableAreaLocationSequenceComponent(BaseModel):
    """Labware on an addressable area."""

    kind: Literal["onAddressableArea"] = "onAddressableArea"
    addressableAreaName: str


class OnCutoutFixtureLocationSequenceComponent(BaseModel):
    """Something on a deck cutout fixture."""

    kind: Literal["onCutoutFixture"] = "onCutoutFixture"
    possibleCutoutFixtureIds: list[str]
    cutoutId: str


class NotOnDeckLocationSequenceComponent(BaseModel):
    """Labware on a system location."""

    kind: Literal["notOnDeck"] = "notOnDeck"
    logicalLocationName: _OffDeckLocationType | _SystemLocationType


LabwareLocationSequence = list[
    OnLabwareLocationSequenceComponent
    | OnModuleLocationSequenceComponent
    | OnAddressableAreaLocationSequenceComponent
    | NotOnDeckLocationSequenceComponent
    | OnCutoutFixtureLocationSequenceComponent
    | InStackerHopperLocation
]
"""Labware location specifier."""

LabwareLocation = Union[
    DeckSlotLocation,
    ModuleLocation,
    OnLabwareLocation,
    _OffDeckLocationType,
    _SystemLocationType,
    AddressableAreaLocation,
    InStackerHopperLocation,
]
"""Union of all locations where it's legal to keep a labware."""

LoadableLabwareLocation = Union[
    DeckSlotLocation,
    ModuleLocation,
    OnLabwareLocation,
    _OffDeckLocationType,
    _SystemLocationType,
    AddressableAreaLocation,
]
"""Union of all locations where it's legal to load a labware."""

OnDeckLabwareLocation = Union[
    DeckSlotLocation, ModuleLocation, OnLabwareLocation, AddressableAreaLocation
]

NonStackedLocation = Union[
    DeckSlotLocation,
    AddressableAreaLocation,
    ModuleLocation,
    _OffDeckLocationType,
    _SystemLocationType,
]
"""Union of all locations where it's legal to keep a labware that can't be stacked on another labware"""


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class DeckPoint(BaseModel):
    """Coordinates of a point in deck space."""

    x: float
    y: float
    z: float
